import logging

import tenacity
from tenacity.wait import wait_base


log = logging.getLogger(__name__)

_RETRY_CODES = [429, 502, 503, 504]
_RETRY_VERBS = ["HEAD", "TRACE", "GET", "PUT", "OPTIONS", "DELETE"]
_POST_RETRY_CODES = [429, 503]


# Defining the default tenacity.Retrying as a user-friendly code string
# so that it can be shown in civis.APIClient's docstring.
DEFAULT_RETRYING_STR = """
tenacity.Retrying(
    wait=tenacity.wait_random_exponential(multiplier=2, max=60),
    stop=(tenacity.stop_after_delay(600) | tenacity.stop_after_attempt(10)),
    retry_error_callback=lambda retry_state: retry_state.outcome.result(),
    reraise=True,
)
"""


def get_default_retrying():
    """Return a new instance of the default tenacity.Retrying."""
    # Explicitly set the available globals and locals
    # to mitigate risk of unwanted code execution
    return eval(  # nosec
        DEFAULT_RETRYING_STR,
        {"tenacity": tenacity, "__builtins__": {}},  # globals
        {},  # locals
    )


def retry_request(method, prepared_req, session, retrying=None):
    retry_conditions = None

    # New tenacity.Retrying instance needed, whether it's a copy of the user-provided
    # one or it's one based on civis-python's default settings.
    retrying = retrying.copy() if retrying else get_default_retrying()

    # If retries are exhausted,
    # raise the last exception encountered, not tenacity's RetryError.
    retrying.reraise = True

    def _make_request(req, sess):
        """send the prepared session request"""
        response = sess.send(req)
        return response

    if method.upper() == "POST":
        retry_conditions = tenacity.retry_if_result(
            lambda res: res.status_code in _POST_RETRY_CODES
        )
    elif method.upper() in _RETRY_VERBS:
        retry_conditions = tenacity.retry_if_result(
            lambda res: res.status_code in _RETRY_CODES
        )

    if retry_conditions:
        retrying.retry = retry_conditions
        retrying.wait = wait_at_least_retry_after_header(base=retrying.wait)
        response = retrying(_make_request, prepared_req, session)
        return response

    response = _make_request(prepared_req, session)
    return response


class wait_at_least_retry_after_header(wait_base):
    """Wait strategy for at least `Retry-After` seconds (if present from header)"""

    def __init__(self, base):
        self.base = base

    def __call__(self, retry_state):
        # retry_state is an instance of tenacity.RetryCallState.
        # The .outcome property contains the result/exception
        # that came from the underlying function.
        headers = retry_state.outcome._result.headers

        try:
            retry_after = float(
                headers.get("Retry-After") or headers.get("retry-after") or "0.0"
            )
        except (TypeError, ValueError):
            retry_after = 0.0
        # Wait at least retry_after seconds (compared to the user-specified wait).
        return max(retry_after, self.base(retry_state))



from civis._camel_to_snake import camel_to_snake


def test_camel_to_snake():
    test_cases = [
        ("CAMELCase", "camel_case"),
        ("camelCase", "camel_case"),
        ("CamelCase", "camel_case"),
        ("c__amel", "c__amel"),
    ]
    for in_word, out_word in test_cases:
        assert camel_to_snake(in_word) == out_word



"""Validation for Civis Platform workflow definitions."""

from __future__ import annotations

import io

import jsonschema
import yaml

from ._schemas import WORKFLOW_SCHEMA


_TASK_TRANSITION_ENGINE_COMMANDS = frozenset(["pause", "succeed", "fail"])


class WorkflowValidationError(Exception):
    pass


def validate_workflow_yaml(wf_def: str, /) -> None:
    """Validate a YAML-formatted workflow definition.

    Parameters
    ----------
    wf_def : str
        YAML-formatted workflow definition.

    Raises
    ------
    WorkflowValidationError
        If the workflow definition is invalid.
    """
    _validate_workflow_yaml_ascii_only(wf_def)
    wf_def_dict = yaml.safe_load(wf_def)
    _validate_workflow_by_schema(wf_def_dict)
    _validate_workflow_tasks(wf_def_dict)


def _validate_workflow_by_schema(wf: dict) -> None:
    try:
        jsonschema.validate(wf, WORKFLOW_SCHEMA)
    except jsonschema.ValidationError as e:
        raise WorkflowValidationError(e)


def _validate_workflow_yaml_ascii_only(wf_def: str) -> None:
    for line_no, line in enumerate(io.StringIO(wf_def), 1):
        for char_no, char in enumerate(line, 1):
            if not char.isascii():
                raise WorkflowValidationError(
                    "Workflow definition YAML cannot contain non-ASCII characters: "
                    f"(line {line_no}) {line!r}, (character {char_no}) {char!r}"
                )


def _get_next_task_names(next_tasks: str | list | dict | None) -> list[str]:
    """Next task names under {'on-success', 'on-error', 'on-complete'}."""
    if next_tasks is None:
        return []
    elif isinstance(next_tasks, str):
        return [next_tasks]
    elif isinstance(next_tasks, list):
        task_names = []
        for task in next_tasks:
            if isinstance(task, str):
                task_names.append(task)
            elif isinstance(task, dict):
                task_names.append(list(task.keys())[0])
            else:
                raise WorkflowValidationError(
                    "each item in next task list must be either of type str or dict: "
                    f"{type(task)} ({task})"
                )
        return task_names
    elif isinstance(next_tasks, dict):
        return _get_next_task_names(next_tasks.get("next"))


def _validate_workflow_tasks(wf: dict) -> None:
    """Custom checks for workflow tasks that aren't amenable to jsonschema validation"""
    key = None
    for k in wf.keys():
        if k != "version":
            key = k
            break
    try:
        tasks = wf[key]["tasks"]
    except KeyError:
        raise WorkflowValidationError("No workflow tasks found")
    for task_name, task in tasks.items():
        recognized_task_names = set(tasks.keys()) | _TASK_TRANSITION_ENGINE_COMMANDS
        for next_task_group_name in ("on-success", "on-error", "on-complete"):
            next_task_names = _get_next_task_names(task.get(next_task_group_name))
            if task_name in next_task_names:
                raise WorkflowValidationError(
                    "A task cannot transition to itself. "
                    f"{task_name!r} transitions to itself in {next_task_group_name!r}."
                )
            for next_task_name in next_task_names:
                if next_task_name not in recognized_task_names:
                    raise WorkflowValidationError(
                        f"Task {task_name!r} transitions "
                        f"to an undefined task {next_task_name!r} "
                        f"in {next_task_group_name!r}."
                    )



"""Test the `civis.polling` module"""

import time
from concurrent import futures
import unittest
from unittest import mock

from civis.response import Response
from civis.polling import PollableResult, _ResultPollingThread

import pytest


class State:
    def __init__(self, state):
        self.state = state


def create_pollable_result(state, exception=None, result=None):
    f = PollableResult(State, (state,), polling_interval=0.001)
    f._exception = exception
    f._result = result
    return f


CANCELLED_RESULT = create_pollable_result(state="cancelled")
FINISHED_RESULT = create_pollable_result(state="success")
QUEUED_RESULT = create_pollable_result(state="queued")


class TestPolling(unittest.TestCase):
    def test_as_completed(self):
        my_futures = [QUEUED_RESULT, CANCELLED_RESULT, FINISHED_RESULT]
        fs = futures.as_completed(my_futures)
        f1 = next(fs)
        f2 = next(fs)
        finished_futures = set([f1, f2])

        self.assertEqual(finished_futures, set([FINISHED_RESULT, CANCELLED_RESULT]))

    def test_wait(self):
        done, not_done = futures.wait(
            [QUEUED_RESULT, FINISHED_RESULT], return_when=futures.FIRST_COMPLETED
        )
        self.assertEqual(set([FINISHED_RESULT]), done)
        self.assertEqual(set([QUEUED_RESULT]), not_done)

    def test_error_passthrough(self):
        pollable = PollableResult(
            mock.Mock(side_effect=[ZeroDivisionError()]), (), polling_interval=0.1
        )
        pytest.raises(ZeroDivisionError, pollable.result)

    def test_error_setting(self):
        pollable = PollableResult(
            mock.Mock(side_effect=[ZeroDivisionError()]), (), polling_interval=0.1
        )
        assert isinstance(pollable.exception(), ZeroDivisionError)

    def test_timeout(self):
        pollable = PollableResult(
            mock.Mock(return_value=Response({"state": "running"})),
            poller_args=(),
            polling_interval=0.1,
        )
        pytest.raises(futures.TimeoutError, pollable.result, timeout=0.05)

    def test_poll_on_creation(self):
        poller = mock.Mock(return_value=Response({"state": "running"}))
        pollable = PollableResult(
            poller, (), polling_interval=0.01, poll_on_creation=False
        )
        pollable.done()  # Check status once to start the polling thread
        assert poller.call_count == 0
        time.sleep(0.05)
        assert poller.call_count > 0

    def test_poller_returns_none(self):
        check_result = mock.Mock(
            side_effect=[None, None, Response({"state": "success"})]
        )
        pollable_result = mock.Mock()
        pollable_result._check_result = check_result
        pollable_result._next_polling_interval = 0.01
        polling_thread = _ResultPollingThread(pollable_result)
        polling_thread.run()
        assert check_result.call_count == 3

    def test_reset_polling_thread(self):
        pollable = PollableResult(
            mock.Mock(return_value=Response({"state": "running"})),
            poller_args=(),
            polling_interval=0.1,
        )
        initial_polling_thread = pollable._polling_thread
        assert pollable.polling_interval == 0.1
        assert pollable._next_polling_interval == 0.1
        pollable._reset_polling_thread(0.2)
        # Check that the polling interval was updated
        assert pollable.polling_interval == 0.2
        assert pollable._next_polling_interval == 0.2
        # Check that the _polling_thread is a new thread
        assert pollable._polling_thread != initial_polling_thread
        # Check that the old thread was stopped
        assert not initial_polling_thread.is_alive()

    def test_geometric_polling(self):
        # To test polling, we make the poller function spit out a timestamp every time
        # it is called. Then we check if these timestamps are what we'd expect.
        poller_timestamps = []

        def append_new_timestamp(*args, **kwargs):
            nonlocal poller_timestamps  # noqa: F824
            poller_timestamps.append(time.time())
            if len(poller_timestamps) < 5:
                return Response({"state": "running"})
            else:
                return Response({"state": "succeeded"})

        poller = mock.Mock()
        poller.side_effect = append_new_timestamp

        pollable = PollableResult(poller, (), poll_on_creation=False)
        start_time = time.time()
        pollable.result()

        assert len(poller_timestamps) == 5
        expected_intervals = [1, 1.2, 1.44, 1.728, 2.0736]
        actual_intervals = []
        for i, timestamp in enumerate(poller_timestamps):
            actual_intervals.append(
                timestamp - (poller_timestamps[i - 1] if i else start_time)
            )
        assert actual_intervals == pytest.approx(expected_intervals, abs=0.02)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import collections
from functools import lru_cache
import logging
import textwrap
import warnings

import tenacity

import civis
from civis.resources import generate_classes_maybe_cached
from civis._get_api_key import get_api_key
from civis._retries import DEFAULT_RETRYING_STR
from civis.response import _RETURN_TYPES, find, find_one


_log = logging.getLogger(__name__)


class APIClient:
    """The Civis API client.

    Parameters
    ----------
    api_key : str, optional
        Your API key obtained from the Civis Platform. If not given, the
        client will use the :envvar:`CIVIS_API_KEY` environment variable.
    return_type : str, optional
        The following types are implemented:

        - ``'raw'`` Returns the raw :class:`requests:requests.Response` object.
        - ``'snake'`` Returns a :class:`civis.Response` object for the
          json-encoded content of a response. This maps the top-level json
          keys to snake_case.
    api_version : string, optional
        The version of endpoints to call. May instantiate multiple client
        objects with different versions. Currently only "1.0" is supported.
    local_api_spec : collections.OrderedDict or string, optional
        The methods on this class are dynamically built from the Civis API
        specification, which can be retrieved from the /endpoints endpoint.
        When local_api_spec is None, the default, this specification is
        downloaded the first time APIClient is instantiated. Alternatively,
        a local cache of the specification may be passed as either an
        OrderedDict or a filename which points to a json file.
    force_refresh_api_spec : bool, optional
        Whether to force re-downloading the API spec,
        even if the cached version for the given API key hasn't expired.
    retries : tenacity.Retrying, optional
        Provide a :class:`tenacity.Retrying` instance for retries.
        If ``None`` or not provided, the following default will be used:

        .. code-block:: python

           {default_retrying}

        If you're providing a :class:`tenacity.Retrying` instance,
        please note that you should leave the ``retry`` attribute unspecified,
        because the conditions under which retries apply are pre-determined
        -- see :ref:`retries` for details.
        In addition, the ``reraise`` attribute will be overridden to ``True``
        to raise the last exception (rather than tenacity's `RetryError`)
        if all retry attempts are exhausted.
    user_agent : str, optional
        A custom user agent string to use for requests made by this client.
        The user agent string will be appended with the Python version,
        Civis Python client version, and requests package version.
    """

    def __init__(
        self,
        api_key: str | None = None,
        return_type: str = "snake",
        api_version: str = "1.0",
        local_api_spec: collections.OrderedDict | str | None = None,
        force_refresh_api_spec: bool = False,
        retries: tenacity.Retrying | None = None,
        user_agent: str | None = None,
    ):
        if return_type not in _RETURN_TYPES:
            raise ValueError(
                f"Return type must be one of {set(_RETURN_TYPES)}: " f"{return_type}"
            )
        self._feature_flags = ()
        session_auth_key = get_api_key(api_key)
        if retries is not None and not isinstance(retries, tenacity.Retrying):
            raise TypeError(
                "If provided, the `retries` parameter must be "
                "a tenacity.Retrying instance."
            )
        self._session_kwargs = {
            "api_key": session_auth_key,
            "retrying": retries,
            "user_agent": user_agent,
        }
        self.last_response = None

        classes = generate_classes_maybe_cached(
            local_api_spec,
            session_auth_key,
            api_version,
            force_refresh_api_spec,
            user_agent,
        )
        for class_name, cls in classes.items():
            setattr(
                self,
                class_name,
                cls(self._session_kwargs, client=self, return_type=return_type),
            )

    @property
    def feature_flags(self):
        """The feature flags for the current user.

        .. deprecated:: 2.6.0
            This property is deprecated and will be removed at civis-python v3.0.0.
            Please use ``client.users.get_me()["feature_flags"]`` instead.

        Returns
        -------
        tuple[str]
        """
        warnings.warn(
            "The property `feature_flags` is deprecated and will be removed "
            "at civis-python v3.0.0. Please use "
            "client.users.get_me()['feature_flags'] instead.",
            FutureWarning,
            stacklevel=2,  # Point to the user code that calls this method.
        )
        if self._feature_flags:
            return self._feature_flags
        me = self.users.get_me()
        self._feature_flags = tuple(
            flag for flag, value in me["feature_flags"].items() if value
        )
        return self._feature_flags

    def __getstate__(self):
        raise RuntimeError("The APIClient object can't be pickled.")

    @lru_cache(maxsize=128)
    def get_database_id(self, database):
        """Return the database ID for a given database name.

        Parameters
        ----------
        database : str or int
            If an integer ID is given, passes through. If a str is given
            the database ID corresponding to that database name is returned.

        Returns
        -------
        database_id : int
            The ID of the database.

        Raises
        ------
        ValueError
            If the database can't be found.
        """
        if isinstance(database, int):
            return database
        db = find_one(self.databases.list(), name=database)
        if not db:
            raise ValueError("Database {} not found.".format(database))

        return db["id"]

    @lru_cache(maxsize=128)
    def get_database_credential_id(self, username, database_name):
        """Return the credential ID for a given username in a given database.

        .. deprecated:: 2.4.0
            This method is deprecated and will be removed at civis-python v3.0.0.
            Its continued usage is strongly discouraged. Given the way Civis
            Platform has evolved over the years, there's currently no reliable
            way to get a database credential ID from a username and database name.
            No replacement for this method is being planned. If you need to
            programmatically access a database credential ID that is or may likely
            be the default credential, consider the property
            :py:attr:`~civis.APIClient.default_database_credential_id`.

        Parameters
        ----------
        username : str or int
            If an integer ID is given, this passes through directly. If a
            str is given, return the ID corresponding to the database
            credential with that username.
        database_name : str or int
            Return the ID of the database credential with username
            `username` for this database name or ID.

        Returns
        -------
        database_credential_id : int
            The ID of the database credentials.

        Raises
        ------
        ValueError
            If the credential can't be found.

        Examples
        --------
        >>> import civis
        >>> client = civis.APIClient()
        >>> client.get_database_credential_id('jsmith', 'redshift-general')
        1234

        >>> client.get_database_credential_id(1111, 'redshift-general')
        1111
        """
        warnings.warn(
            "The method `get_database_credential_id` is deprecated and will be removed "
            "at civis-python v3.0.0. Its continued usage is strongly discouraged. "
            "Given the way Civis Platform has evolved over the years, "
            "there's currently no reliable way to get a database credential ID "
            "from a username and database name. No replacement for this method is "
            "being planned. If you need to programmatically access a database "
            "credential ID that is or may likely be the default credential, "
            "consider the property `default_database_credential_id`.",
            FutureWarning,
            stacklevel=2,  # Point to the user code that calls this method.
        )
        if isinstance(username, int):
            return username
        else:
            creds = self.credentials.list(type="Database")
            filter_kwargs = {"username": username}
            if isinstance(database_name, int):
                filter_kwargs["remote_host_id"] = database_name
            else:
                filter_kwargs["remote_host_name"] = database_name
            my_creds = find_one(creds, **filter_kwargs)
            if my_creds is None:
                raise ValueError(
                    "Credential ID for {} on {} not "
                    "found.".format(username, database_name)
                )

        return my_creds["id"]

    @lru_cache(maxsize=128)
    def get_aws_credential_id(self, cred_name, owner=None):
        """Find an AWS credential ID.

        Parameters
        ----------
        cred_name : str or int
            If an integer ID is given, this passes through directly. If a
            str is given, return the ID corresponding to the AWS credential
            with that name.
        owner : str, optional
            Return the credential with this owner. If not provided, search
            for credentials under your username to disambiguate multiple
            credentials with the same name. Note that this function cannot
            return credentials which are not associated with an owner.

        Returns
        -------
        aws_credential_id : int
            The ID number of the AWS credentials.

        Raises
        ------
        ValueError
            If the AWS credential can't be found.

        Examples
        --------
        >>> import civis
        >>> client = civis.APIClient()
        >>> client.get_aws_credential_id('jsmith')
        1234

        >>> client.get_aws_credential_id(1111)
        1111

        >>> client.get_aws_credential_id('shared-cred',
        ...                              owner='research-group')
        99
        """
        if isinstance(cred_name, int):
            return cred_name
        else:
            creds = self.credentials.list(type="Amazon Web Services S3")
            my_creds = find(creds, name=cred_name)
            if owner is not None:
                my_creds = find(my_creds, owner=owner)

            if not my_creds:
                own_str = "" if owner is None else " owned by {}".format(owner)
                msg = "AWS credential ID for {}{} cannot be found"
                raise ValueError(msg.format(cred_name, own_str))
            elif len(my_creds) > 1:
                if owner is None:
                    # If the user didn't specify an owner, see if we can
                    # narrow down to just credentials owned by this user.
                    owner = self.username
                    my_creds = find(my_creds, owner=owner)
                if len(my_creds) > 1:
                    _log.warning(
                        "Found %d AWS credentials with name %s and "
                        "owner %s. Returning the first.",
                        len(my_creds),
                        cred_name,
                        owner,
                    )
            my_creds = my_creds[0]

        return my_creds["id"]

    @lru_cache(maxsize=128)
    def get_table_id(self, table, database):
        """Return the table ID for a given database and table name.

        Parameters
        ----------
        table : str
            The name of the table in format schema.tablename.
            Either schema or tablename, or both, can be double-quoted to
            correctly parse special characters (such as '.').
        database : str or int
            The name or ID of the database.

        Returns
        -------
        table_id : int
            The ID of the table.

        Raises
        ------
        ValueError
            If a table match can't be found.

        Examples
        --------
        >>> import civis
        >>> client = civis.APIClient()
        >>> client.get_table_id('foo.bar', 'redshift-general')
        123
        >>> client.get_table_id('"schema.has.periods".bar', 'redshift-general')
        456
        """
        database_id = self.get_database_id(database)
        schema, name = civis.io.split_schema_tablename(table)
        tables = self.tables.list(database_id=database_id, schema=schema, name=name)
        if not tables:
            msg = "No tables found for {} in database {}"
            raise ValueError(msg.format(table, database))

        return tables[0].id

    @lru_cache(maxsize=128)
    def get_storage_host_id(self, storage_host):
        """Return the storage host ID for a given storage host name.

        Parameters
        ----------
        storage_host : str or int
            If an integer ID is given, passes through. If a str is given
            the storage host ID corresponding to that storage host is returned.

        Returns
        -------
        storage_host_id : int
            The ID of the storage host.

        Raises
        ------
        ValueError
            If the storage host can't be found.

        Examples
        --------
        >>> import civis
        >>> client = civis.APIClient()
        >>> client.get_storage_host_id('test host')
        1234

        >>> client.get_storage_host_id(1111)
        1111
        """
        if isinstance(storage_host, int):
            return storage_host
        sh = find_one(self.storage_hosts.list(), name=storage_host)
        if not sh:
            raise ValueError("Storage Host {} not found.".format(storage_host))

        return sh["id"]

    @property
    @lru_cache(maxsize=128)
    def default_credential(self):
        """The current user's default credential.

        .. deprecated:: 2.4.0
            This property is deprecated and will be removed at civis-python v3.0.0.
            Please use :py:attr:`~civis.APIClient.default_database_credential_id`
            instead.
        """
        warnings.warn(
            "The property `default_credential` is deprecated and will be removed "
            "at civis-python v3.0.0. "
            "Please use `default_database_credential_id` instead.",
            FutureWarning,
            stacklevel=2,  # Point to the user code that calls this method.
        )
        creds = self.credentials.list(default=True)
        return creds[0]["id"] if len(creds) > 0 else None

    @property
    @lru_cache(maxsize=128)
    def default_database_credential_id(self):
        """The current user's default database credential ID."""
        creds = self.credentials.list(
            default=True, type="Database", remote_host_id=None
        )
        return creds[0]["id"] if len(creds) > 0 else None

    @property
    @lru_cache(maxsize=128)
    def username(self):
        """The current user's username."""
        return self.users.get_me().username


APIClient.__doc__ = APIClient.__doc__.format(
    default_retrying=textwrap.indent(DEFAULT_RETRYING_STR.strip(), " " * 11).strip()
)



from __future__ import annotations

import json
import pprint
from collections.abc import Callable, Iterable, Iterator

from typing import Any, Generic, TypeVar

import requests

from civis._camel_to_snake import camel_to_snake


_RETURN_TYPES = frozenset({"snake", "raw"})

# "arguments": Script arguments are often environment variables in
# ALL_CAPS that we don't want to convert to snake_case.
# "environmentVariables": from objects under the `Services` endpoint
_RESPONSE_KEYS_PRESERVE_CASE = frozenset({"arguments", "environmentVariables"})


class CivisClientError(Exception):
    def __init__(self, message: str, response: requests.Response):
        self.status_code = response.status_code
        self.error_message = message

    def __str__(self):
        return self.error_message


class CivisImmutableResponseError(Exception):
    pass


def _response_to_json(response):
    """Parse a raw response to a dict.

    Parameters
    ----------
    response: requests.Response
        A raw response returned by an API call.

    Returns
    -------
    dict | None
        The data in the response body or None if the response has no
        content.

    Raises
    ------
    CivisClientError
        If the data in the raw response cannot be parsed.
    """
    if response.content == b"":
        return None
    else:
        try:
            return response.json()
        except ValueError:
            raise CivisClientError("Unable to parse JSON from response", response)


def convert_response_data_type(
    response, headers=None, return_type="snake", from_json_values=False
):
    """Convert a raw response into a given type.

    Parameters
    ----------
    response : list, dict, or `requests.Response`
        Convert this object into a different response object.
    headers : dict, optional
        If given and the return type supports it, attach these headers to the
        converted response. If `response` is a `requests.Response`, the headers
        will be inferred from it.
    return_type : string, {'snake', 'raw'}
        Convert the response to this type. See documentation on
        `civis.APIClient` for details of the return types.
    from_json_values : bool, optional
        If True, the `response` comes from the `json_values` endpoint.

    Returns
    -------
    dict, `civis.Response`, `civis.ListResponse`, or `requests.Response`
        Depending on the value of `return_type`.
    """
    if return_type == "raw":
        return response

    elif return_type == "snake":
        if isinstance(response, requests.Response):
            headers = response.headers
            data = _response_to_json(response)
        else:
            data = response

        if isinstance(data, list):
            responses = [
                Response(d, headers=headers, from_json_values=from_json_values)
                for d in data
            ]
            return ListResponse(responses, headers)
        else:
            return Response(data, headers=headers, from_json_values=from_json_values)

    else:
        raise ValueError(f"Return type not one of {set(_RETURN_TYPES)}: {return_type}")


def _raise_response_immutable_error():
    raise CivisImmutableResponseError(
        "Response object is read-only. "
        "Did you want to call .json() for a dictionary that you can modify?"
    )


class Response:
    """Custom Civis response object.

    Attributes
    ----------
    json_data : dict | None
        This is `json_data` as it is originally returned to the user without
        the key names being changed. None is used if the original
        response returned a 204 No Content response.
    headers : dict
        This is the header for the API call without changing the key names.
    calls_remaining : int
        Number of API calls remaining before rate limit is reached.
    rate_limit : int
        Total number of calls per API rate limit period.
    """

    def __init__(
        self,
        json_data: dict[str, Any] | None,
        *,
        headers: dict | None = None,
        snake_case: bool = True,
        from_json_values: bool = False,
    ):
        self.json_data = json_data
        self.headers = headers
        self.calls_remaining = (
            int(x) if (x := (headers or {}).get("X-RateLimit-Remaining")) else x
        )
        self.rate_limit = (
            int(x) if (x := (headers or {}).get("X-RateLimit-Limit")) else x
        )

        # Note that these two dicts can have Response objects as values.
        self._data_camel = {}
        self._data_snake = {}

        if json_data is not None:
            for key, v in json_data.items():

                if key == "value" and (
                    from_json_values or json_data.get("objectType") == "JSONValue"
                ):
                    # When json_data represents a JSONValue (either from one of the
                    # methods under the `json_values` endpoint or from a script's run
                    # output), `v` is the deserialized JSON.
                    val = v
                elif isinstance(v, dict):
                    if key in _RESPONSE_KEYS_PRESERVE_CASE:
                        val = Response(v, snake_case=False)
                    else:
                        val = Response(v)
                elif isinstance(v, list):
                    val = [Response(o) if isinstance(o, dict) else o for o in v]
                else:
                    val = v

                key_snake = camel_to_snake(key) if snake_case else key

                self._data_camel[key] = val
                self._data_snake[key_snake] = val

    def json(self, snake_case: bool = True) -> dict[str, Any]:
        """Return the JSON data.

        Parameters
        ----------
        snake_case : bool, optional
            If True (the default), return the keys in snake case.
            If False, return the keys in camel case.

        Returns
        -------
        dict
        """
        if self.json_data is None:
            return {}
        elif snake_case:
            return self._to_dict_with_snake_case_keys()
        else:
            return self.json_data.copy()

    def _to_dict_with_snake_case_keys(self):
        result = {}
        for k, v in self._data_snake.items():
            if isinstance(v, list):
                result[k] = [
                    o._to_dict_with_snake_case_keys() if isinstance(o, Response) else o
                    for o in v
                ]
            elif isinstance(v, Response):
                result[k] = v._to_dict_with_snake_case_keys()
            else:
                result[k] = v
        return result

    def __setattr__(self, key, value):
        if key == "__dict__":
            self.__dict__.update(value)
        elif key in (
            "json_data",
            "headers",
            "calls_remaining",
            "rate_limit",
            "_data_camel",
            "_data_snake",
        ):
            self.__dict__[key] = value
        else:
            _raise_response_immutable_error()

    def __setitem__(self, key, value):
        _raise_response_immutable_error()

    def __getitem__(self, item):
        try:
            return self._data_snake[item]
        except KeyError:
            return self._data_camel[item]

    def __getattr__(self, item):
        try:
            return self.__getitem__(item)
        except KeyError:
            raise AttributeError(f"Response object has no attribute {item!r}")

    def __len__(self):
        return len(self._data_snake)

    def __repr__(self):
        return f"Response({repr(self._data_snake)})"

    def __hash__(self):
        return hash(json.dumps(self.json_data))

    def _repr_pretty_(self, p, cycle):
        """Pretty-print the response object in IPython and Jupyter.

        https://ipython.readthedocs.io/en/stable/api/generated/IPython.lib.pretty.html#extending
        """
        if cycle:
            p.text("Response(...)")
        else:
            p.text(pprint.pformat(self))

    def get(self, key, default=None):
        """Get the value for the given key."""
        try:
            return self.__getitem__(key)
        except KeyError:
            return default

    def items(self):
        """Return an iterator of the key-value pairs in the response."""
        return self._data_snake.items()

    def __eq__(self, other):
        if isinstance(other, dict):
            return self._data_snake == other
        elif isinstance(other, Response):
            return self._data_snake == other._data_snake
        else:
            return False

    def __setstate__(self, state):
        """Set the state when unpickling, to avoid RecursionError."""
        self.__dict__ = state


class _safe_key:
    """Helper function for key functions when sorting unorderable objects.

    The wrapped-object will fallback to a Py2.x style comparison for
    unorderable types (sorting first comparing the type name and then by
    the obj ids).  Does not work recursively, so dict.items() must have
    _safe_key applied to both the key and the value.

    Source: https://github.com/python/cpython/blob/3.13/Lib/pprint.py#L80-L100
    """

    __slots__ = ["obj"]

    def __init__(self, obj):
        self.obj = obj

    def __lt__(self, other):
        try:
            return self.obj < other.obj
        except TypeError:
            return (str(type(self.obj)), id(self.obj)) < (
                str(type(other.obj)),
                id(other.obj),
            )


def _safe_tuple(t):
    """Helper function for comparing 2-tuples

    Source: https://github.com/python/cpython/blob/3.13/Lib/pprint.py#L102-L104
    """
    return _safe_key(t[0]), _safe_key(t[1])


def _pprint_response(self, object, stream, indent, allowance, context, level):
    """Pretty-print a Response object.

    Inspired by https://stackoverflow.com/a/52521743
    Based on python's dict pprint:
    https://github.com/python/cpython/blob/3.7/Lib/pprint.py#L180-L192
    """
    write = stream.write
    object = object._data_snake
    write("Response({")
    if self._indent_per_level > 1:
        write((self._indent_per_level - 1) * " ")
    length = len(object)
    if length:
        if self._sort_dicts:
            items = sorted(object.items(), key=_safe_tuple)
        else:
            items = object.items()
        # The 9 in `indent + 9` is the length of "Response(".
        self._format_dict_items(
            items, stream, indent + 9, allowance + 1, context, level
        )
    write("})")


pprint.PrettyPrinter._dispatch[Response.__repr__] = _pprint_response  # type: ignore

T_Response = TypeVar("T_Response", bound=Response)


class PaginatedResponse(Generic[T_Response]):
    """A generator of :class:`civis.Response` objects, for paginated API calls.

    Parameters
    ----------
    path : str
        Make GET requests to this path.
    initial_params : dict
        Query params that should be passed along with each request. Note that
        if `initial_params` contains the key `page_num`, it will
        be ignored. The given dict is not modified.
    endpoint : `civis.base.Endpoint`
        An endpoint used to make API requests.

    Notes
    -----
    This response is returned automatically by endpoints which support
    pagination when the `iterator` kwarg is specified.

    Examples
    --------
    >>> import civis
    >>> client = civis.APIClient()
    >>> queries = client.queries.list(iterator=True)
    >>> for query in queries:
    ...    print(query['id'])
    """

    def __init__(self, path: str, initial_params: dict[str, Any], endpoint):
        self._path = path
        self._params = initial_params.copy()
        self._endpoint = endpoint

        # We are paginating through all items, so start at the beginning.
        self._params["page_num"] = 1

        self._iter: Iterator[T_Response] | None = None

    def __iter__(self) -> Iterator[T_Response]:
        return self

    def _get_iter(self) -> Iterator[T_Response]:
        while True:
            response = self._endpoint._make_request("GET", self._path, self._params)
            page_data = _response_to_json(response)
            if len(page_data) == 0:
                return

            for data in page_data:
                converted_data = convert_response_data_type(
                    data,
                    headers=response.headers,
                    return_type=self._endpoint._return_type,
                    from_json_values=(self._path or "").startswith("json_values"),
                )
                yield converted_data

            self._params["page_num"] += 1

    def __next__(self) -> T_Response:
        if self._iter is None:
            self._iter = self._get_iter()
        return next(self._iter)

    def json(self, snake_case: bool = True) -> list[dict[str, Any]]:
        """Return the JSON data of all responses.

        Parameters
        ----------
        snake_case : bool, optional
            If True (the default), return the keys in snake case.
            If False, return the keys in camel case.

        Returns
        -------
        list[dict]
        """
        if self._iter is None:
            self._iter = self._get_iter()
        return [r.json(snake_case) for r in self._iter]


class ListResponse(list[T_Response], Generic[T_Response]):
    """A list of :class:`civis.Response` objects.

    Parameters
    ----------
    responses : list[civis.Response]
        A list of response objects to be stored in this list.
    headers : dict, optional
        Headers to be attached to the list response.
        The headers info is available as an attribute, so that it can be accessed
        even when the list is empty.
    """

    __slots__ = ("headers",)

    def __init__(self, responses: list[T_Response], headers: dict | None = None):
        super().__init__(responses)
        self.headers = headers

    def json(self, snake_case: bool = True) -> list[dict[str, Any]]:
        """Return the JSON data of all responses in the list.

        Parameters
        ----------
        snake_case : bool, optional
            If True (the default), return the keys in snake case.
            If False, return the keys in camel case.

        Returns
        -------
        list[dict]
        """
        return [r.json(snake_case) for r in self]


def find(
    object_list: Iterable[Response], filter_func: Callable | None = None, **kwargs
) -> list[Response]:
    """Filter :class:`civis.Response` objects.

    Parameters
    ----------
    object_list : iterable
        An iterable of arbitrary objects, particularly those with attributes
        that can be targeted by the filters in `kwargs`. A major use case is
        an iterable of :class:`civis.Response` objects.
    filter_func : callable, optional
        A one-argument function. If specified, `kwargs` are ignored.
        An `object` from the input iterable is kept in the returned list
        if and only if ``bool(filter_func(object))`` is ``True``.
    **kwargs
        Key-value pairs for more fine-grained filtering; they cannot be used
        in conjunction with ``filter_func``. All keys must be strings.
        For an object ``obj`` from the input iterable to be included in the
        returned list, all the keys must be attributes of ``obj``, plus
        any one of the following conditions for a given key:

        - ``value`` is a one-argument function and
          ``bool(value(getattr(obj, key)))`` is equal to ``True``
        - ``value`` is either ``True`` or ``False``, and
          ``getattr(obj, key) is value`` is ``True``
        - ``getattr(obj, key) == value`` is ``True``

    Returns
    -------
    list

    Examples
    --------
    >>> import civis
    >>> client = civis.APIClient()
    >>> # creds is a list of civis.Response objects
    >>> creds = client.credentials.list()
    >>> # target_creds contains civis.Response objects
    >>> # with the attribute 'name' == 'username'
    >>> target_creds = find(creds, name='username')

    See Also
    --------
    civis.find_one
    """
    if filter_func is None:

        def filter_func(o):
            for k, v in kwargs.items():
                if not hasattr(o, k):
                    return False
                elif callable(v):
                    if not v(getattr(o, k, None)):
                        return False
                elif isinstance(v, bool):
                    if getattr(o, k) is not v:
                        return False
                elif v != getattr(o, k, None):
                    return False
            return True

    return [o for o in object_list if filter_func(o)]


def find_one(
    object_list: Iterable[Response], filter_func: Callable | None = None, **kwargs
) -> Response | None:
    """Return one satisfying :class:`civis.Response` object.

    The arguments are the same as those for :func:`civis.find`.
    If more than one object satisfies the filtering criteria,
    the first one is returned.
    If no satisfying objects are found, ``None`` is returned.

    Returns
    -------
    object or None

    See Also
    --------
    civis.find
    """
    results = find(object_list, filter_func, **kwargs)

    return results[0] if results else None



import json
import warnings
from operator import itemgetter
from unittest import mock

import pytest
import requests

from civis import APIClient, response
from civis.base import CivisAPIError, CivisJobFailure
from civis.futures import (
    ContainerFuture,
    _ContainerShellExecutor,
    CustomScriptExecutor,
    _create_docker_command,
)

from civis.futures import CivisFuture
from civis.tests import create_client_mock, create_client_mock_for_container_tests


def _create_poller_mock(state: str) -> mock.Mock:
    api_result = mock.Mock(state=state)
    poller = mock.Mock(return_value=api_result)
    return poller


def test_check_message():
    mock_civis = create_client_mock()
    result = CivisFuture(lambda x: x, (1, 20), client=mock_civis)
    message = {"object": {"id": 1}, "run": {"id": 20, "state": "succeeded"}}
    assert result._check_message(message) is True


def test_check_message_with_different_run_id():
    mock_civis = create_client_mock()
    result = CivisFuture(lambda x: x, (1, 20), client=mock_civis)
    message = {"object": {"id": 2}, "run": {"id": 20, "state": "succeeded"}}
    assert result._check_message(message) is False


def test_check_message_when_job_is_running():
    mock_civis = create_client_mock()
    result = CivisFuture(lambda x: x, (1, 20), client=mock_civis)
    message = {"object": {"id": 1}, "run": {"id": 20, "state": "running"}}
    assert result._check_message(message) is False


def test_poller_call_count_poll_on_creation_true():
    mock_civis = create_client_mock()
    poller = _create_poller_mock("succeeded")
    CivisFuture(poller, (1, 2), poll_on_creation=True, client=mock_civis)
    assert poller.call_count == 1


def test_poller_call_count_poll_on_creation_false():
    mock_civis = create_client_mock()
    poller = _create_poller_mock("succeeded")
    CivisFuture(poller, (1, 2), poll_on_creation=False, client=mock_civis)
    assert poller.call_count == 0


def test_set_api_result_succeeded():
    mock_civis = create_client_mock()
    poller = _create_poller_mock("succeeded")
    result = CivisFuture(poller, (1, 2), client=mock_civis)
    assert result._state == "FINISHED"


@mock.patch("civis.futures.time.sleep", side_effect=lambda x: None)
def test_set_api_result_failed(m_sleep):
    mock_civis = create_client_mock()
    poller = _create_poller_mock("failed")
    result = CivisFuture(poller, (1, 2), client=mock_civis)
    assert result._state == "FINISHED"
    with pytest.raises(CivisJobFailure):
        result.result()
    with pytest.raises(CivisJobFailure):
        result.outputs()


def test_outputs_succeeded():
    poller = _create_poller_mock("succeeded")
    mock_client = create_client_mock()
    expected_return = [{"test": "test_result"}]
    mock_client.jobs.list_runs_outputs.return_value = expected_return

    result = CivisFuture(poller, (1, 2), client=mock_client)
    assert result.outputs() == expected_return


def test_polling_interval():
    mock_client = create_client_mock()
    polling_interval = 30
    future = CivisFuture(
        lambda x: x, (1, 20), polling_interval=polling_interval, client=mock_client
    )
    assert future.polling_interval == polling_interval


def _check_executor(from_template_id=None):
    job_id, run_id = 42, 43
    c = _setup_client_mock(job_id, run_id, n_failures=0)
    mock_run = c.scripts.post_containers_runs()
    if from_template_id:
        bpe = CustomScriptExecutor(
            from_template_id=from_template_id, client=c, polling_interval=0.01
        )
        future = bpe.submit(my_param="spam")
    else:
        bpe = _ContainerShellExecutor(client=c, polling_interval=0.01)
        future = bpe.submit("foo")

    # Mock and test running, future.job_id, and done()
    mock_run_json = mock_run.json()
    mock_run_json["state"] = "running"
    mock_run = response.Response(mock_run_json)
    assert future.running(), "future is incorrectly marked as not running"
    assert future.job_id == job_id, "job_id not stored properly"
    assert not future.done(), "future is incorrectly marked as done"

    future.cancel()

    # Mock and test cancelled()
    assert future.cancelled(), "cancelled() did not return True as expected"
    assert not future.running(), "running() did not return False as expected"

    # Mock and test done()
    mock_run_json = mock_run.json()
    mock_run_json["state"] = "succeeded"
    mock_run = response.Response(mock_run_json)
    assert future.done(), "done() did not return True as expected"

    # Test cancelling all jobs.
    mock_run_json = mock_run.json()
    mock_run_json["state"] = "running"
    mock_run = response.Response(mock_run_json)
    bpe.cancel_all()
    assert future.cancelled(), "cancel_all() failed"

    # Test shutdown method.
    bpe.shutdown(wait=True)
    assert future.done(), "shutdown() failed"

    return c


@pytest.mark.parametrize(
    "poller_args,expected_job_id,expected_run_id",
    [((123, 456), 123, 456), ((123,), 123, None)],
)
def test_future_job_id_run_id(poller_args, expected_job_id, expected_run_id):
    result = CivisFuture(
        poller=_create_poller_mock("succeeded"),
        poller_args=poller_args,
        client=create_client_mock(),
    )
    assert result.job_id == expected_job_id
    assert result.run_id == expected_run_id
    assert result.job_url == (
        f"https://platform.civisanalytics.com/spa/#/jobs/{expected_job_id}"
    )


def test_container_future_job_id_run_id():
    job_id, run_id = 123, 456
    result = ContainerFuture(
        job_id=job_id,
        run_id=run_id,
        client=create_client_mock_for_container_tests(),
    )
    assert result.job_id == job_id
    assert result.run_id == run_id
    assert result.job_url == f"https://platform.civisanalytics.com/spa/#/jobs/{job_id}"


def test_container_scripts():
    c = _check_executor()
    assert c.scripts.post_custom.call_count == 0
    assert c.scripts.post_containers.call_count > 0


def test_custom_scripts():
    with mock.patch.dict("os.environ", {"CIVIS_JOB_ID": "12", "CIVIS_RUN_ID": "40"}):
        c = _check_executor(133)
    assert c.scripts.post_custom.call_count > 0
    assert c.scripts.post_containers.call_count == 0

    # Verify that this script's job and run ID are passed to arguments
    args = c.scripts.post_custom.call_args[1].get("arguments")
    for k, v in (("CIVIS_PARENT_JOB_ID", "12"), ("CIVIS_PARENT_RUN_ID", "40")):
        assert args.get(k) == v


@pytest.mark.parametrize("is_child_job", [False, True])
def test_container_script_param_injection(is_child_job):
    # Test that child jobs created by the shell executor have the
    # job and run IDs of the script which created them (if any).
    job_id, run_id = "123", "13"
    c = _setup_client_mock(42, 43, n_failures=0)
    mock_env = {"CIVIS_JOB_ID": job_id, "CIVIS_RUN_ID": run_id}

    with mock.patch.dict("os.environ", mock_env):
        init_kwargs = dict(client=c, polling_interval=0.01)
        if is_child_job:
            init_kwargs["params"] = [
                {"name": "CIVIS_PARENT_JOB_ID", "type": "integer", "value": "888"},
                {"name": "CIVIS_PARENT_RUN_ID", "type": "integer", "value": "999"},
            ]
        bpe = _ContainerShellExecutor(**init_kwargs)
        bpe.submit("foo")

    params = sorted(
        c.scripts.post_containers.call_args[1].get("params"), key=itemgetter("name")
    )
    assert params == [
        {"name": "CIVIS_PARENT_JOB_ID", "type": "integer", "value": job_id},
        {"name": "CIVIS_PARENT_RUN_ID", "type": "integer", "value": run_id},
    ], "The parent job parameters were not set correctly."


def test_create_docker_command():
    res = _create_docker_command(
        "foo.sh", "bar", "baz", wibble="wibble1", wobble="wobble1"
    )
    assert res == "foo.sh bar baz --wibble wibble1 --wobble wobble1"


# A function to raise fake API errors the first
# num_failures times it is called.
def _make_error_func(num_failures, failure_is_error=False):
    """Raise API errors multiple times before succeeding

    Test error-handling code by using this to mock
    calls to post_containers_runs or get_containers_runs.

    Parameters
    ----------
    num_failures: int
        Fail this many times before returning a success
    failure_is_error: bool
        If True, "failure" means raising a `CivisAPIError`.

    Returns
    -------
    MockRun
        Mock which imitates the result of a `post_containers_runs`
        or `get_containers_runs` call
    """
    counter = {"failures": 0}  # Use a dict so we can modify it in the closure

    def mock_api_error(job_id, run_id):
        if counter["failures"] < num_failures:
            counter["failures"] += 1
            if failure_is_error:
                raise CivisAPIError(mock.MagicMock())
            else:
                return response.Response(
                    {"id": run_id, "container_id": job_id, "state": "failed"}
                )
        else:
            return response.Response(
                {"id": run_id, "container_id": job_id, "state": "succeeded"}
            )

    return mock_api_error


def _setup_client_mock(job_id=-10, run_id=100, n_failures=8, failure_is_error=False):
    """Return a Mock set up for use in testing container scripts

    Parameters
    ----------
    job_id: int
        Mock-create containers with this ID when calling `post_containers`
        or `post_containers_runs`.
    run_id: int
        Mock-create runs with this ID when calling `post_containers_runs`.
    n_failures: int
        When calling `get_containers_runs`, fail this many times
        before succeeding.
    failure_is_error: bool
        If True, "failure" means raising a `CivisAPIError`.

    Returns
    -------
    `unittest.mock.Mock`
        With `post_containers`, `post_containers_runs`, and
        `get_containers_runs` methods set up.
    """
    c = mock.Mock()
    c.__class__ = APIClient

    mock_container = response.Response({"id": job_id})
    c.scripts.post_containers.return_value = mock_container
    c.scripts.post_custom.return_value = mock_container
    mock_container_run = response.Response(
        {"id": run_id, "container_id": job_id, "state": "queued"}
    )
    c.scripts.post_containers_runs.return_value = mock_container_run
    c.jobs.post_runs.return_value = mock_container_run
    c.scripts.get_containers_runs.side_effect = _make_error_func(
        n_failures, failure_is_error
    )

    def change_state_to_cancelled(job_id):
        mock_container_run_json = mock_container_run.json()
        mock_container_run_json["state"] = "cancelled"
        return response.Response(mock_container_run_json)

    c.scripts.post_cancel.side_effect = change_state_to_cancelled

    return c


def test_cancel_finished_job():
    # If we try to cancel a completed job, we get a 404 error.
    # That shouldn't be sent to the user.

    # Set up a mock client which will give an exception when
    # you try to cancel any job.
    c = _setup_client_mock()
    err_resp = requests.Response()
    status_code = 404
    err_resp.status_code = status_code
    err_resp._content = json.dumps(
        {
            "status_code": status_code,
            "error": "not_found",
            "errorDescription": "The requested resource could not be found.",
            "content": True,
        }
    ).encode()
    c.scripts.post_cancel.side_effect = CivisAPIError(err_resp)
    resp_json = c.scripts.post_containers_runs.return_value.json()
    resp_json["state"] = "running"
    c.scripts.post_containers_runs.return_value = response.Response(resp_json)
    fut = ContainerFuture(
        -10, 100, polling_interval=1, client=c, poll_on_creation=False
    )
    assert not fut.done()
    with warnings.catch_warnings():
        # Check that no warnings are raised.
        warnings.simplefilter("error")
        assert fut.cancel() is False


def test_future_no_retry_error():
    # Verify that with no retries, exceptions on job polling
    #  are raised to the user
    c = _setup_client_mock(failure_is_error=True)
    fut = ContainerFuture(-10, 100, polling_interval=0.001, client=c)
    with pytest.raises(CivisAPIError):
        fut.result()


def test_future_no_retry_failure():
    # Verify that with no retries, job failures are raised as
    # exceptions for the user
    c = _setup_client_mock(failure_is_error=False)
    fut = ContainerFuture(-10, 100, polling_interval=0.001, client=c)
    with pytest.raises(CivisJobFailure):
        fut.result()


def test_future_not_enough_retry_error():
    # Verify that if polling the run is still erroring after all retries
    # are exhausted, the error will be raised for the user.
    c = _setup_client_mock(failure_is_error=True)
    fut = ContainerFuture(-10, 100, max_n_retries=3, polling_interval=0.01, client=c)
    with pytest.raises(CivisAPIError):
        fut.result()


def test_future_not_enough_retry_failure():
    # Verify that if the job is still failing after all retries
    # are exhausted, the job failure will be raised for the user.
    c = _setup_client_mock(failure_is_error=False)
    fut = ContainerFuture(-10, 100, max_n_retries=3, polling_interval=0.01, client=c)
    with pytest.raises(CivisJobFailure):
        fut.result()


def test_future_retry_failure():
    # Verify that we can retry through API errors until a job succeeds
    c = _setup_client_mock(failure_is_error=False)
    fut = ContainerFuture(-10, 100, max_n_retries=10, polling_interval=0.01, client=c)
    assert fut.result().state == "succeeded"


def test_future_retry_error():
    # Verify that we can retry through job failures until it succeeds
    c = _setup_client_mock(failure_is_error=True)
    fut = ContainerFuture(-10, 100, max_n_retries=10, polling_interval=0.01, client=c)
    assert fut.result().state == "succeeded"


@mock.patch("civis.futures.time.sleep", side_effect=lambda x: None)
def test_container_exception_no_result_logs(m_sleep):
    # If the job errored with no output but with logs,
    # we should return error logs with the future exception.
    mem_msg = "Run used approximately 2 millicores " "of its 256 millicore CPU limit"
    failed_msg = "Failed: The job container failed. Exit code 1"
    logs = [
        {"id": 111, "created_at": "abc", "message": mem_msg, "level": "info"},
        {"id": 222, "created_at": "def", "message": failed_msg, "level": "error"},
    ]
    mock_client = create_client_mock_for_container_tests(
        1, 2, state="failed", run_outputs=[], log_outputs=logs
    )
    fut = ContainerFuture(1, 2, client=mock_client)

    with pytest.raises(CivisJobFailure) as err:
        fut.result()
    expected_msg = "(From job 1 / run 2) " + "\n".join([failed_msg, mem_msg, ""])
    assert expected_msg == str(fut._exception.error_message)
    assert str(err.value) == expected_msg


@mock.patch("civis.futures.time.sleep", side_effect=lambda x: None)
def test_container_exception_memory_error(m_sleep):
    err_msg = "Process ran out of its allowed 3000 MiB of " "memory and was killed."
    logs = [
        {
            "created_at": "2017-05-10T12:00:00.000Z",
            "id": 10005,
            "level": "error",
            "message": "Failed",
        },
        {
            "created_at": "2017-05-10T12:00:00.000Z",
            "id": 10003,
            "level": "error",
            "message": "Error on job: Process ended with an " "error, exiting: 137.",
        },
        {
            "created_at": "2017-05-10T12:00:00.000Z",
            "id": 10000,
            "level": "error",
            "message": err_msg,
        },
    ]
    mock_client = create_client_mock_for_container_tests(
        1, 2, state="failed", run_outputs=[], log_outputs=logs
    )
    fut = ContainerFuture(1, 2, client=mock_client)

    with pytest.raises(MemoryError) as err:
        fut.result()
    assert str(err.value) == f"(From job 1 / run 2) {err_msg}"



"""
This is an executable intended for use with a joblib backend
for the Civis platform. It takes a Civis File ID representing
a callable serialized by either ``pickle`` or ``cloudpickle``
as an argument, downloads the file,
deserializes it, calls the callable, serializes the result,
and uploads the result to another Civis File. The output file's ID
will be set as an output on this run.
"""

from datetime import datetime, timedelta
from io import BytesIO
import os
import pickle  # nosec
import sys

import civis
import cloudpickle
from joblib import parallel_config

from civis.parallel import (
    _robust_pickle_download,
    _robust_file_to_civis,
    _setup_remote_backend,
)


def worker_func(func_file_id):
    # Have the output File expire in 7 days.
    expires_at = (datetime.now() + timedelta(days=7)).isoformat()

    client = civis.APIClient()
    job_id = os.environ.get("CIVIS_JOB_ID")
    run_id = os.environ.get("CIVIS_RUN_ID")
    if not job_id or not run_id:
        raise RuntimeError("This function must be run inside a " "Civis container job.")

    # Run the function.
    result = None
    try:
        func, remote_backend = _robust_pickle_download(
            func_file_id, client=client, n_retries=5, delay=0.5
        )

        _backend = _setup_remote_backend(remote_backend)

        with parallel_config(_backend):
            result = func()
    except Exception as exc:
        print("Error! Attempting to record exception.")
        result = exc
        raise
    finally:
        # Serialize the result and upload it to the Files API.
        if result is not None:
            # If the function exits without erroring, we may not have a result.
            result_buffer = BytesIO()
            cloudpickle.dump(result, result_buffer, pickle.HIGHEST_PROTOCOL)
            result_buffer.seek(0)
            output_name = "Results from Joblib job {} / run {}".format(job_id, run_id)
            output_file_id = _robust_file_to_civis(
                result_buffer,
                output_name,
                n_retries=5,
                delay=0.5,
                expires_at=expires_at,
                client=client,
            )
            client.scripts.post_containers_runs_outputs(
                job_id, run_id, "File", output_file_id
            )
            print("Results output to file ID: {}".format(output_file_id))


def main():
    if len(sys.argv) > 1:
        func_file_id = sys.argv[1]
    else:
        # If the file ID to download isn't given as a command-line
        # argument, assume that it's in an environment variable.
        func_file_id = os.environ["JOBLIB_FUNC_FILE_ID"]
    worker_func(func_file_id=func_file_id)


if __name__ == "__main__":
    main()



import concurrent.futures as cf
from datetime import datetime
from math import floor
from unittest import mock

import pytest
import tenacity
from requests import Request
from requests import ConnectionError

from civis._retries import retry_request, get_default_retrying
from civis._retries import _RETRY_VERBS, _RETRY_CODES, _POST_RETRY_CODES


def _get_retrying(retries: int):
    retrying = get_default_retrying()
    stop = tenacity.stop_after_delay(600) | tenacity.stop_after_attempt(retries)
    retrying.stop = stop
    return retrying


def test_no_retry_on_success():
    expected_call_count = 0
    api_response = {"key": "value"}
    mock_session = mock.MagicMock()
    session_context = mock_session.return_value.__enter__.return_value
    session_context.send.return_value.json.return_value = api_response

    for verb in _RETRY_VERBS:
        expected_call_count += 1
        session_context.send.return_value.status_code = 200

        request_info = dict(
            params={"secondParameter": "b", "firstParameter": "a"},
            json={},
            url="https://api.civisanalytics.com/wobble/wubble",
            method=verb,
        )
        request = Request(**request_info)
        pre_request = session_context.prepare_request(request)
        retry_request(verb, pre_request, session_context, _get_retrying(3))

        assert session_context.send.call_count == expected_call_count


def test_no_retry_on_get_no_retry_failure():
    expected_call_count = 0
    max_calls = 3
    api_response = {"key": "value"}
    mock_session = mock.MagicMock()
    session_context = mock_session.return_value.__enter__.return_value
    session_context.send.return_value.json.return_value = api_response

    for verb in _RETRY_VERBS:
        expected_call_count += 1
        session_context.send.return_value.status_code = 403

        request_info = dict(
            params={"secondParameter": "b", "firstParameter": "a"},
            json={},
            url="https://api.civisanalytics.com/wobble/wubble",
            method=verb,
        )
        request = Request(**request_info)
        pre_request = session_context.prepare_request(request)
        retry_request(verb, pre_request, session_context, _get_retrying(max_calls))

        assert session_context.send.call_count == expected_call_count


@mock.patch("civis.futures.time.sleep", side_effect=lambda x: None)
def test_retry_on_retry_eligible_failures(m_sleep):
    expected_call_count = 0
    max_calls = 3
    api_response = {"key": "value"}
    mock_session = mock.MagicMock()
    session_context = mock_session.return_value.__enter__.return_value
    session_context.send.return_value.json.return_value = api_response
    for verb in _RETRY_VERBS:
        for code in _RETRY_CODES:
            expected_call_count += max_calls
            session_context.send.return_value.status_code = code

            request_info = dict(
                params={"secondParameter": "b", "firstParameter": "a"},
                json={},
                url="https://api.civisanalytics.com/wobble/wubble",
                method=verb,
            )

            request = Request(**request_info)
            pre_request = session_context.prepare_request(request)
            retry_request(verb, pre_request, session_context, _get_retrying(max_calls))

            assert session_context.send.call_count == expected_call_count


@mock.patch("civis.futures.time.sleep", side_effect=lambda x: None)
def test_retry_on_retry_eligible_failures_lowercase_verbs(m_sleep):
    expected_call_count = 0
    max_calls = 3
    api_response = {"key": "value"}
    mock_session = mock.MagicMock()
    session_context = mock_session.return_value.__enter__.return_value
    session_context.send.return_value.json.return_value = api_response
    for verb in _RETRY_VERBS:
        for code in _RETRY_CODES:
            expected_call_count += max_calls
            session_context.send.return_value.status_code = code

            request_info = dict(
                params={"secondParameter": "b", "firstParameter": "a"},
                json={},
                url="https://api.civisanalytics.com/wobble/wubble",
                method=verb.lower(),
            )

            request = Request(**request_info)
            pre_request = session_context.prepare_request(request)
            retry_request(verb, pre_request, session_context, _get_retrying(max_calls))

            assert session_context.send.call_count == expected_call_count


def test_no_retry_on_post_success():
    expected_call_count = 1
    max_calls = 3
    api_response = {"key": "value"}
    mock_session = mock.MagicMock()
    session_context = mock_session.return_value.__enter__.return_value
    session_context.send.return_value.json.return_value = api_response

    session_context.send.return_value.status_code = 200

    request_info = dict(
        params={"secondParameter": "b", "firstParameter": "a"},
        json={},
        url="https://api.civisanalytics.com/wobble/wubble",
        method="POST",
    )
    request = Request(**request_info)
    pre_request = session_context.prepare_request(request)
    retry_request("post", pre_request, session_context, _get_retrying(max_calls))

    assert session_context.send.call_count == expected_call_count


@mock.patch("civis.futures.time.sleep", side_effect=lambda x: None)
def test_retry_on_retry_eligible_post_failures(m_sleep):
    expected_call_count = 0
    max_calls = 3
    api_response = {"key": "value"}
    mock_session = mock.MagicMock()
    session_context = mock_session.return_value.__enter__.return_value
    session_context.send.return_value.json.return_value = api_response

    for code in _POST_RETRY_CODES:
        expected_call_count += max_calls
        session_context.send.return_value.status_code = code

        request_info = dict(
            params={"secondParameter": "b", "firstParameter": "a"},
            json={},
            url="https://api.civisanalytics.com/wobble/wubble",
            method="POST",
        )
        request = Request(**request_info)
        pre_request = session_context.prepare_request(request)
        retry_request("post", pre_request, session_context, _get_retrying(max_calls))

        assert session_context.send.call_count == expected_call_count


def test_no_retry_on_connection_error():
    expected_call_count = 0
    api_response = {"key": "value"}
    mock_session = mock.MagicMock()
    session_context = mock_session.return_value.__enter__.return_value
    session_context.send.return_value.json.return_value = api_response

    for verb in _RETRY_VERBS:
        expected_call_count += 1

        request_info = dict(
            params={"secondParameter": "b", "firstParameter": "a"},
            json={},
            url="https://api.civisanalytics.com/wobble/wubble",
            method=verb,
        )
        request = Request(**request_info)
        pre_request = session_context.prepare_request(request)

        session_context.send.side_effect = ConnectionError()
        try:
            retry_request(verb, pre_request, session_context, _get_retrying(3))
        except ConnectionError:
            pass

        assert session_context.send.call_count == expected_call_count


def _retry_respect_retry_after_headers(verb):
    expected_call_count = 0
    max_calls = 2
    retry_after = 1
    api_response = {"key": "value"}
    mock_session = mock.MagicMock()
    session_context = mock_session.return_value.__enter__.return_value
    session_context.send.return_value.json.return_value = api_response

    session_context.send.return_value.status_code = 429
    session_context.send.return_value.headers = {"Retry-After": str(retry_after)}

    expected_call_count += max_calls

    request_info = dict(
        params={"secondParameter": "b", "firstParameter": "a"},
        json={},
        url="https://api.civisanalytics.com/wobble/wubble",
        method=verb,
    )

    request = Request(**request_info)
    pre_request = session_context.prepare_request(request)

    start_time = datetime.now().timestamp()
    retry_request(verb, pre_request, session_context, _get_retrying(max_calls))
    end_time = datetime.now().timestamp()
    duration = end_time - start_time

    assert session_context.send.call_count == expected_call_count
    assert floor(duration) == retry_after * (max_calls - 1)


def test_retry_respect_retry_after_headers():
    verbs = ["HEAD", "TRACE", "GET", "PUT", "OPTIONS", "DELETE", "POST"]
    all_verbs = [v for v in (verbs + [v.lower() for v in verbs])]
    with cf.ThreadPoolExecutor() as executor:
        executor.map(_retry_respect_retry_after_headers, all_verbs)


@pytest.mark.parametrize(
    "retrying, max_calls",
    [
        # Simulate a user-provided tenacity.Retrying with limited max attempts.
        (_get_retrying(3), 3),
        # No user-provided tenacity.Retrying. Use civis-python's default retrying.
        (None, 10),
    ],
)
@mock.patch("civis.futures.time.sleep", side_effect=lambda x: None)
def test_no_recursion_in_retry_request(m_sleep, retrying, max_calls):
    """Test that retry_request (used inside civis.APIClient) doesn't cause recursion."""

    api_response = {"key": "value"}
    mock_session = mock.MagicMock()
    session_context = mock_session.return_value.__enter__.return_value
    session_context.send.return_value.json.return_value = api_response
    session_context.send.return_value.status_code = 429
    session_context.send.return_value.headers = {}

    # Simulate multiple API calls using the same `retrying` object.
    iterations = 1001  # 1001 consecutive API calls to trigger deep recursion
    for i in range(iterations):
        request_info = dict(
            params={"call_number": str(i)},
            json={},
            url="https://api.civisanalytics.com/test",
            method="GET",
        )
        request = Request(**request_info)
        pre_request = session_context.prepare_request(request)

        # This should not cause recursion even after multiple calls
        retry_request("GET", pre_request, session_context, retrying)

    # Verify that all calls completed without recursion error
    # Each call should have attempted max_calls times
    assert session_context.send.call_count == max_calls * iterations



from __future__ import annotations

import os
import platform
import threading
from concurrent import futures
from json.decoder import JSONDecodeError
from posixpath import join

import requests

import civis
from civis.response import PaginatedResponse, convert_response_data_type, Response
from civis._retries import retry_request

FINISHED = ["success", "succeeded"]
FAILED = ["failed"]
NOT_FINISHED = ["queued", "running"]
CANCELLED = ["cancelled"]
DONE = FINISHED + FAILED + CANCELLED

# Translate Civis state strings into `future` state strings
STATE_TRANS = {}
for name in FINISHED + FAILED:
    STATE_TRANS[name] = futures._base.FINISHED
for name in NOT_FINISHED:
    STATE_TRANS[name] = futures._base.RUNNING
for name in CANCELLED:
    STATE_TRANS[name] = futures._base.CANCELLED_AND_NOTIFIED


DEFAULT_API_ENDPOINT = "https://api.civisanalytics.com/"

DEFAULT_USER_AGENT = "civis-python"


def tostr_urljoin(*x):
    return join(*map(str, x))


class CivisJobFailure(Exception):
    def __init__(
        self,
        err_msg: str,
        response: Response | None = None,
        job_id: int | None = None,
        run_id: int | None = None,
    ):
        self.job_id = job_id
        self.run_id = run_id
        self._original_err_msg = err_msg
        self.error_message = _err_msg_with_job_run_ids(err_msg, job_id, run_id)
        self.response = response

    def __str__(self):
        return self.error_message


def _err_msg_with_job_run_ids(err_msg, job_id, run_id) -> str:
    if job_id is None and run_id is None:
        return str(err_msg)
    elif run_id is None:
        return f"(From job {job_id}) {err_msg}"
    else:
        return f"(From job {job_id} / run {run_id}) {err_msg}"


class CivisAPIError(Exception):
    def __init__(self, response: requests.Response):
        if response.content:  # the API itself gave an error response
            try:
                json = response.json()
            except JSONDecodeError as e:
                if "Expecting value: line 1 column 1 (char 0)" in str(e):
                    self.error_message = "No Response Content from Civis API"
                else:
                    self.error_message = (
                        f"Response Content: " f"{(response.content or b'').decode()}"
                    )
            else:
                self.error_message = (
                    json.get("errorDescription") or "No Error Message Available"
                )
        else:  # this was something like a 502
            self.error_message = response.reason

        self.status_code = response.status_code
        self._response = response

    def __str__(self):
        if self.status_code:
            method = self._response.request.method
            url = self._response.request.path_url
            return f"({self.status_code}) {self.error_message} ({method} {url})"
        else:
            return self.error_message


class EmptyResultError(Exception):
    pass


class CivisAPIKeyError(Exception):
    pass


class CivisImportError(Exception):
    pass


def get_base_url():
    base_url = os.environ.get("CIVIS_API_ENDPOINT", DEFAULT_API_ENDPOINT)
    if not base_url.endswith("/"):
        base_url += "/"
    return base_url


def get_headers(user_agent):
    user_agent = (
        f"{(user_agent or DEFAULT_USER_AGENT).strip()} "
        f"(Python {platform.python_version()}; "
        f"civis-python {civis.__version__}; "
        f"requests {requests.__version__})"
    )
    headers = {"User-Agent": user_agent}
    job_id, run_id = os.getenv("CIVIS_JOB_ID"), os.getenv("CIVIS_RUN_ID")
    if job_id:
        headers.update({"X-Civis-Job-ID": job_id, "X-Civis-Run-ID": run_id})
    return headers


class Endpoint:

    _lock = threading.Lock()

    def __init__(self, session_kwargs, client, return_type="raw"):
        self._session_kwargs = session_kwargs
        self._return_type = return_type
        self._base_url = get_base_url()
        self._client = client
        self._headers = get_headers(self._session_kwargs.get("user_agent", {}))

    def _build_path(self, path):
        if not path:
            return self._base_url
        return tostr_urljoin(self._base_url, path.strip("/"))

    @staticmethod
    def _handle_array_params(params):
        """Convert array-like parameters to the format that Civis API expects.

        Different APIs expect array-like parameters in different formats.
        For Civis API, an array parameter `foo` needs to be passed in as `foo[]`.
        Related reference: https://stackoverflow.com/a/23347265
        """
        if not params:
            return
        new_params = {}
        for key, value in params.items():
            if hasattr(value, "__iter__") and not isinstance(value, (dict, str)):
                new_params[f"{key}[]"] = list(value)
            else:
                new_params[key] = value
        return new_params

    def _make_request(self, method, path=None, params=None, data=None, **kwargs):
        url = self._build_path(path)
        params = self._handle_array_params(params)

        with self._lock:
            with open_session(self._session_kwargs["api_key"], self._headers) as sess:
                request = requests.Request(
                    method, url, json=data, params=params, **kwargs
                )
                pre_request = sess.prepare_request(request)
                response = retry_request(
                    method, pre_request, sess, self._session_kwargs["retrying"]
                )

        if response.status_code == 401:
            auth_error = response.headers["www-authenticate"]
            raise CivisAPIKeyError(auth_error) from CivisAPIError(response)

        if not response.ok:
            raise CivisAPIError(response)

        return response

    def _call_api(
        self,
        method,
        path=None,
        params=None,
        data=None,
        **kwargs,
    ):
        iterator = kwargs.pop("iterator", False)

        if iterator:
            resp = PaginatedResponse(path, params, self)
        else:
            resp = self._make_request(method, path, params, data, **kwargs)
            resp = convert_response_data_type(
                resp,
                return_type=self._return_type,
                from_json_values=(path or "").startswith("json_values"),
            )
        self._client.last_response = resp
        return resp


class CivisAsyncResultBase(futures.Future):
    """A base class for tracking asynchronous results.

    Sub-classes needs to call either the `set_result` method to set a result
    when it finishes or call `set_exception` if there is an error.

    The `_result` attribute can also be set to change the current state of
    the result. The `_result_` object needs to be set to an object with a
    ``state`` attribute. Alternatively the `_check_result` method can be
    overwritten to change how the state of the object is returned.
    """

    def __repr__(self):
        # Almost the same as the superclass's __repr__, except we use
        # the `_civis_state` rather than the `_state`.
        with self._condition:
            if self._civis_state in FINISHED + FAILED:
                if self.exception():
                    return "<%s at %#x state=%s raised %s>" % (
                        self.__class__.__name__,
                        id(self),
                        self._civis_state,
                        self._exception.__class__.__name__,
                    )
                else:
                    return "<%s at %#x state=%s returned %s>" % (
                        self.__class__.__name__,
                        id(self),
                        self._civis_state,
                        self.result().__class__.__name__,
                    )
            out = "<%s at %#x state=%s>" % (
                self.__class__.__name__,
                id(self),
                self._civis_state,
            )
            return out

    def cancel(self):
        """Not currently implemented."""
        raise NotImplementedError("Running jobs cannot currently be cancelled")

    def succeeded(self):
        """Return ``True`` if the job completed in Civis with no error."""
        with self._condition:
            return self._civis_state in FINISHED

    def failed(self):
        """Return ``True`` if the Civis job failed."""
        with self._condition:
            return self._civis_state in FAILED

    def _check_result(self):
        with self._condition:
            if self._result is not None:
                return self._result

    @property
    def _civis_state(self):
        """State as returned from Civis."""
        with self._condition:
            if self._check_result():
                return self._check_result().state
            return "running"

    @property
    def _state(self):
        """State of the CivisAsyncResultBase in `future` language."""
        with self._condition:
            return STATE_TRANS[self._civis_state]

    @_state.setter
    def _state(self, value):
        # Ignore attempts to set the _state from the `Future` superclass
        pass

    def set_result(self, result):
        """Sets the return value of work associated with the future.

        This is adapted from
        https://github.com/python/cpython/blob/3.8/Lib/concurrent/futures/_base.py#L517-L530
        This version does not try to change the _state or check that the
        initial _state is running since the Civis implementation has _state
        depend on the Platform job state.
        """
        with self._condition:
            self._result = result
            for waiter in self._waiters:
                waiter.add_result(self)
            self._condition.notify_all()
        self._invoke_callbacks()

    def set_exception(self, exception):
        """Sets the result of the future as being the given exception.

        This is adapted from
        https://github.com/python/cpython/blob/3.8/Lib/concurrent/futures/_base.py#L532-L545
        This version does not try to change the _state or check that the
        initial _state is running since the Civis implementation has _state
        depend on the Platform job state.
        """
        with self._condition:
            self._exception = exception
            for waiter in self._waiters:
                waiter.add_exception(self)
            self._condition.notify_all()
        self._invoke_callbacks()


def open_session(api_key, headers=None):
    """Create a new Session which can connect with the Civis API"""
    session = requests.Session()
    session.auth = (api_key, "")
    session.headers.update(headers or {})
    return session



import importlib
import sys
from importlib.metadata import version
from typing import TYPE_CHECKING

from civis.client import APIClient
from civis.loggers import civis_logger
from civis.response import find, find_one, ListResponse, Response, PaginatedResponse
from civis.service_client import ServiceClient


try:
    # __sphinx_build__ is injected into bulitins in docs/conf.py.
    if __sphinx_build__:  # type: ignore[name-defined]
        _BUILDING_SPHINX_DOC = True
except NameError:
    _BUILDING_SPHINX_DOC = False

__version__ = version("civis")


def _lazy_import(name):
    # https://docs.python.org/3/library/importlib.html#implementing-lazy-imports
    spec = importlib.util.find_spec(name)
    loader = importlib.util.LazyLoader(spec.loader)
    spec.loader = loader
    module = importlib.util.module_from_spec(spec)
    sys.modules[name] = module
    loader.exec_module(module)
    return module


if TYPE_CHECKING or _BUILDING_SPHINX_DOC:
    # Regularly loaded modules are needed for typing information and doc generation.
    from civis import futures, io, ml, parallel, utils, workflows
else:
    futures = _lazy_import("civis.futures")
    io = _lazy_import("civis.io")
    ml = _lazy_import("civis.ml")
    parallel = _lazy_import("civis.parallel")
    utils = _lazy_import("civis.utils")
    workflows = _lazy_import("civis.workflows")


__all__ = [
    "__version__",
    "APIClient",
    "civis_logger",
    "find",
    "find_one",
    "futures",
    "io",
    "ListResponse",
    "ml",
    "PaginatedResponse",
    "parallel",
    "Response",
    "ServiceClient",
    "utils",
    "workflows",
]



from __future__ import annotations

import os
import uuid
from typing import TypeAlias


TypePathLike: TypeAlias = str | bytes | os.PathLike


def maybe_get_random_name(name):
    if not name:
        name = uuid.uuid4().hex
    return name



import json
import warnings
from collections import OrderedDict
from unittest import mock

import pytest

from civis import APIClient
from civis.resources import API_SPEC_PATH


with open(API_SPEC_PATH) as f:
    API_SPEC = json.load(f, object_pairs_hook=OrderedDict)


class FakeUsersEndpoint:
    def get_me(self):
        return {"feature_flags": {"foo": True, "bar": True, "baz": False}}


@mock.patch("civis.resources._resources.get_api_spec", return_value=API_SPEC)
def test_feature_flags(mock_spec):
    client = APIClient()
    setattr(client, "users", FakeUsersEndpoint())

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        assert client.feature_flags == ("foo", "bar")


@mock.patch("civis.resources._resources.get_api_spec", return_value=API_SPEC)
def test_feature_flags_memoized(mock_spec):
    client = APIClient()
    setattr(client, "users", FakeUsersEndpoint())
    with warnings.catch_warnings():
        with mock.patch.object(client.users, "get_me", wraps=client.users.get_me):
            warnings.simplefilter("ignore")
            client.feature_flags
            client.feature_flags
            assert client.users.get_me.call_count == 1


def test_get_table_id():
    client = APIClient(local_api_spec=API_SPEC, api_key="none")
    client.get_database_id = mock.Mock(return_value=123)

    mock_tables = mock.MagicMock()
    mock_tables.__getitem__.side_effect = {0: mock.Mock()}.__getitem__

    client.tables.list = mock.Mock(return_value=mock_tables)

    client.get_table_id(table="foo.bar", database=123)

    client.tables.list.assert_called_once_with(
        database_id=123, schema="foo", name="bar"
    )


def test_get_storage_host_id():
    client = APIClient(local_api_spec=API_SPEC, api_key="none")

    class StorageHost:
        def __init__(self, id, name):
            self.id = id
            self.name = name

        def __getitem__(self, key):
            return getattr(self, key)

    storage_hosts = [StorageHost(1234, "test"), StorageHost(5678, "othertest")]
    client.storage_hosts.list = mock.Mock(return_value=storage_hosts)

    assert client.get_storage_host_id("test") == 1234

    client.storage_hosts.list.assert_called_once_with()

    assert client.get_storage_host_id(4732) == 4732
    with pytest.raises(ValueError, match="Storage Host invalidname not found"):
        client.get_storage_host_id("invalidname")



from collections import OrderedDict
import json
import os
import shutil
import warnings
from unittest import mock

import pytest

from civis.cli.__main__ import generate_cli, invoke, make_operation_name
from civis.cli._cli_commands import _str_table_result
from civis.resources import API_SPEC_PATH

THIS_DIR = os.path.dirname(os.path.realpath(__file__))


def test_civis_command_available():
    command = "civis"
    assert shutil.which(command), f"The `{command}` command is not available."


@mock.patch("civis.cli.__main__.add_extra_commands")
@mock.patch("civis.cli.__main__.retrieve_spec_dict")
def test_generate_cli_petstore(mock_retrieve_spec_dict, mock_add_extra_commands):
    """Test loading the OpenAPI petstore example."""

    # From https://raw.githubusercontent.com/OAI/OpenAPI-Specification/4b1c1167b99844fd3ca19dc0055bbdb0c5eff094/examples/v2.0/json/petstore.json  # noqa: E501
    with open(os.path.join(THIS_DIR, "petstore.json")) as f:
        petstore_spec = json.load(f)

    mock_retrieve_spec_dict.return_value = petstore_spec
    cli = generate_cli()
    assert set(cli.commands.keys()) == {"pets"}
    assert set(cli.commands["pets"].commands.keys()) == {"list", "post", "get"}
    assert {x.name for x in cli.commands["pets"].commands["list"].params} == {
        "limit",
        "json_output",
    }


@mock.patch("civis.cli.__main__.retrieve_spec_dict")
def test_generate_cli_civis(mock_retrieve_spec_dict):
    """Test loading the Civis API spec as of 2021-12-02."""
    with open(API_SPEC_PATH) as f:
        civis_spec = json.load(f, object_pairs_hook=OrderedDict)
    mock_retrieve_spec_dict.return_value = civis_spec

    with warnings.catch_warnings():
        warnings.simplefilter("error")
        cli = generate_cli()

    # Check a regular command.
    list_runs_cmd = cli.commands["scripts"].commands["list-containers-runs"]
    expected_names = {"id", "limit", "page_num", "order", "order_dir", "json_output"}
    assert {_.name for _ in list_runs_cmd.params} == expected_names
    assert list_runs_cmd.params[0].name == "id"
    assert list_runs_cmd.params[0].required
    assert list_runs_cmd.params[1].name == "limit"
    assert not list_runs_cmd.params[1].required

    # Check that the extra files upload command was added
    expected_names = {"path", "name", "expires_at", "description"}
    files_upload_params = cli.commands["files"].commands["upload"].params
    assert {_.name for _ in files_upload_params} == expected_names

    # Check that the POST queries command, which uses an object in the body,
    # was parsed properly.
    pq_params = cli.commands["queries"].commands["post"].params
    expected_names = {
        "column_delimiter",
        "compression",
        "credential",
        "database",
        "filename_prefix",
        "include_header",
        "interactive",
        "preview_rows",
        "sql",
        "unquoted",
        "json_output",
        "hidden",
    }
    assert {_.name for _ in pq_params} == expected_names
    for p in pq_params:
        if p.name == "filename_prefix":
            assert not p.required
        if p.name == "database":
            assert p.required


@mock.patch("civis.cli.__main__.open_session", autospec=True)
def test_blank_output(mock_session):
    """
    Test that endpoints that return blank results don't cause exceptions.
    """
    # The response object's json method will raise a ValueError when the output
    # is blank.
    session_context = mock_session.return_value.__enter__.return_value
    session_context.send.return_value.json.side_effect = ValueError()
    session_context.send.return_value.status_code = 200

    op = {"parameters": []}
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        invoke("WIBBLE", "/wobble/wubble", op)

    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 0


@mock.patch("civis.cli.__main__.open_session", autospec=True)
def test_failure_exit_code(mock_session):
    """
    Test that we return a nonzero exit code when the API request fails.
    """
    # first test that we get a zero exit code when the API request succeeds
    session_context = mock_session.return_value.__enter__.return_value
    session_context.send.return_value.json.side_effect = ValueError()
    session_context.send.return_value.status_code = 200

    op = {"parameters": []}

    with pytest.raises(SystemExit) as pytest_wrapped_e:
        invoke("WIBBLE", "/wobble/wubble", op)
    assert pytest_wrapped_e.value.code == 0

    # now test that we get a nonzero exit code when the API request fails
    session_context.send.return_value.status_code = 404

    with pytest.raises(SystemExit) as pytest_wrapped_e:
        invoke("WIBBLE", "/wobble/wubble", op)
    assert pytest_wrapped_e.value.code != 0


@mock.patch("civis.cli.__main__.open_session", autospec=True)
@mock.patch("civis.cli.__main__.Request", autospec=True)
def test_parameter_case(mock_request, mock_session):
    """
    Test that parameter names are sent in camelCase rather than snake_case.
    """
    api_response = {"key": "value"}
    session_context = mock_session.return_value.__enter__.return_value
    session_context.send.return_value.json.return_value = api_response
    session_context.send.return_value.status_code = 200

    # To avoid needing CIVIS_API_KEY set in the environment.
    op = {
        "parameters": [
            {"name": "firstParameter", "in": "query"},
            {"name": "secondParameter", "in": "query"},
        ]
    }
    with pytest.raises(SystemExit):
        invoke(
            "WIBBLE", "/wobble/wubble", op, first_parameter="a", second_parameter="b"
        )

    mock_session.call_args[1]["user_agent"] = "civis-cli"

    mock_request.assert_called_with(
        url="https://api.civisanalytics.com/wobble/wubble",
        json={},
        params={"firstParameter": "a", "secondParameter": "b"},
        method="WIBBLE",
    )


@pytest.mark.parametrize(
    "path,method,resource_name,exp",
    [
        ("/imports/files/{id}/runs/{run_id}", "get", "imports", "get-files-runs"),
        ("/aliases/{object_type}/{alias}", "get", "aliases", "get-object-type"),
        ("/workflows/", "get", "workflows", "list"),
        ("/results/{id}/grants", "delete", "results", "delete-grants"),
    ],
)
def test_make_operation_name(path, method, resource_name, exp):
    assert make_operation_name(path, method, resource_name) == exp


def test_str_table_result():
    cols = ["a", "snake!"]
    rows = [["2", "3"], ["1.1", None]]

    out = _str_table_result(cols, rows)
    assert out == "a   | snake!\n------------\n  2 |      3\n1.1 |       "



from __future__ import annotations

import collections
import json
import concurrent.futures
import csv
from os import path
import io
import logging
import os
import shutil
from tempfile import TemporaryDirectory
import warnings
import zlib

import gzip
import zipfile

from civis import APIClient
from civis.io._utils import maybe_get_random_name
from civis.base import EmptyResultError, CivisImportError, CivisAPIError
from civis.futures import CivisFuture
from civis.io import civis_to_file, file_to_civis
from civis.utils import run_job
from civis._deprecation import DeprecatedKwargDefault
from civis._retries import get_default_retrying

import requests

try:
    import pandas as pd

    HAS_PANDAS = True
except ImportError:
    HAS_PANDAS = False

try:
    import polars as pl

    HAS_POLARS = True
except ImportError:
    HAS_POLARS = False

CHUNK_SIZE = 32 * 1024
log = logging.getLogger(__name__)

DELIMITERS = {
    ",": "comma",
    "\t": "tab",
    "|": "pipe",
}

_File = collections.namedtuple("_File", "id name detected_info")

_SQL_PARAMS_ARGUMENTS_KEYS = frozenset(("params", "arguments"))

_RETURN_AS_OPTIONS = frozenset(("list", "pandas", "polars"))


def _validate_return_as(return_as):
    if return_as not in _RETURN_AS_OPTIONS:
        raise ValueError(f"unsupported return_as option: {return_as}")

    if return_as == "pandas" and not HAS_PANDAS:
        raise ImportError("return_as is 'pandas' but pandas is not installed.")
    elif return_as == "polars" and not HAS_POLARS:
        raise ImportError("return_as is 'polars' but polars is not installed.")


def _warn_or_raise_for_use_pandas(use_pandas, return_as):
    """When it's time to actually remove use_pandas at civis-python v3.0.0,
    Just remove this helper function as well as all usage of use_pandas."""

    if not isinstance(use_pandas, DeprecatedKwargDefault):
        warn_msg = (
            "To support multiple dataframe libraries, the keyword argument "
            "'use_pandas' is deprecated since civis-python v2.6.0 "
            "and will be removed in civis-python v3.0.0 "
            "(no release timeline yet). "
            "While 'use_pandas' still works for now, you're strongly encouraged to "
            "update your code to use the new keyword argument 'return_as' instead and "
            "stop settting 'use_pandas'. "
        )
        conflict_msg = (
            "Update your code so that the 'use_pandas' argument is no longer set, "
            "and set 'return_as' to one of {'list', 'pandas', 'polars'}."
        )
        if use_pandas and return_as == "list":
            # return_as has a default value of "list". If return_as is "list",
            # there's no way to tell if the user sets it explicitly or if it's
            # from the default value (unless we either rewrite the function signature
            # or use a decorator, which we can't or don't want to).
            # So when use_pandas=True and return_as="list", assume the user's intent
            # is to return a pandas dataframe.
            return_as = "pandas"
            warn_msg += 'To return a pandas dataframe, set return_as="pandas". '
        elif use_pandas and return_as == "polars":
            raise ValueError(
                "Conflicting argument values: use_pandas=True but return_as='polars'. "
                + conflict_msg
            )
        elif not use_pandas and return_as == "pandas":
            raise ValueError(
                "Conflicting argument values: use_pandas=False but return_as='pandas'. "
                + conflict_msg
            )
        # stacklevel=3 to point the warning to the user's code
        warnings.warn(warn_msg.strip(), FutureWarning, stacklevel=3)
    return return_as


def read_civis(
    table: str,
    database: str | int,
    columns: list[str] | None = None,
    return_as: str = "list",
    use_pandas=DeprecatedKwargDefault(),
    encoding: str | None = None,
    job_name: str | None = None,
    client: APIClient | None = None,
    credential_id: int | None = None,
    polling_interval: int | float | None = None,
    hidden: bool = True,
    **kwargs,
) -> list | pd.DataFrame | pl.DataFrame:
    """Read data from a Civis table.

    Parameters
    ----------
    table : str
        Name of table, including schema, in the database. E.g.
        ``'my_schema.my_table'``. Schemas or tablenames with periods must
        be double quoted, e.g. ``'my_schema."my.table"'``.
    database : str or int
        Read data from this database. Can be the database name or ID.
    columns : list, optional
        A list of column names. Column SQL transformations are possible.
        If omitted, all columns are exported.
    return_as : str, {"list", "pandas", "polars"}
        If ``"list"`` (the default), return a list.
        If ``"pandas"``, return a :class:`pandas.DataFrame`.
        If ``"polars"``, return a :class:`polars.DataFrame`.
    use_pandas : bool, optional
        If ``True``, return a :class:`pandas:pandas.DataFrame`. Otherwise,
        return a list of results from :func:`python:csv.reader`.

        .. deprecated:: 2.6.0
            ``use_pandas`` will be removed at civis-python v3.0.0.
            Please use ``return_as`` instead.
    encoding : str, optional
        If ``return_as`` is ``"pandas"`` or ``"polars"``, this parameter is passed to
        the ``encoding`` kwarg of :func:`pandas.read_csv` or
        :func:`polars.read_csv`, respectively.
        If ``return_as`` is ``"list"``, and if this parameter isn't provided,
        then the UTF-8 encoding is assumed. In case you encounter
        a ``UnicodeDecodeError``, consider choosing an encoding suitable
        for your data; see the `list of standard encodings
        <https://docs.python.org/3/library/codecs.html#standard-encodings>`_.
    job_name : str, optional
        A name to give the job. If omitted, a random job name will be
        used.
    client : :class:`civis.APIClient`, optional
        If not provided, an :class:`civis.APIClient` object will be
        created from the :envvar:`CIVIS_API_KEY`.
    credential_id : str or int, optional
        The database credential ID.  If ``None``, the default credential
        will be used.
    polling_interval : int or float, optional
        Number of seconds to wait between checks for query completion.
    hidden : bool, optional
        If ``True`` (the default), this job will not appear in the Civis UI.
    **kwargs : kwargs
        Extra keyword arguments are passed into
        :func:`pandas.read_csv` if ``return_as`` is ``"pandas"``, or
        passed into :func:`python:csv.reader` if ``return_as`` is
        ``"list"``,
        or passed into :func:`polars.read_csv` if ``return_as`` is ``"polars"``.

    Returns
    -------
    data : list | :class:`pandas.DataFrame` | :class:`polars.DataFrame`
        A list of rows (with header as first row) if ``return_as`` is
        ``"list"``, or :class:`pandas.DataFrame` if ``return_as`` is ``"pandas"``,
        or :class:`polars.DataFrame` if ``return_as`` is ``"polars"``.
        Note that if ``return_as`` is ``"list"``, no parsing of types is performed and
        each row will be a list of strings.

    Raises
    ------
    ImportError
        If ``return_as`` is ``"pandas"`` and pandas is not installed.
        If ``return_as`` is ``"polars"`` and polars is not installed.
    EmptyResultError
        If the table is empty.

    Examples
    --------
    >>> import civis
    >>> table = "schema.table"
    >>> database = "my_data"
    >>> columns = ["column_a", "ROW_NUMBER() OVER(ORDER BY date) AS order"]
    >>> data = civis.io.read_civis(table, database, columns=columns)
    >>> columns = data.pop(0)
    >>> col_a_index = columns.index("column_a")
    >>> col_a = [row[col_a_index] for row in data]

    >>> df = civis.io.read_civis("schema.table", "my_data", return_as="pandas")
    >>> col_a = df["column_a"]

    See Also
    --------
    civis.io.read_civis_sql : Read directly into memory using SQL.
    civis.io.civis_to_csv : Write directly to csv.
    civis.io.export_to_civis_file : Store a SQL query's results in a Civis file
    """
    _validate_return_as(return_as)
    return_as = _warn_or_raise_for_use_pandas(use_pandas, return_as)

    if client is None:
        client = APIClient()
    sql = _get_sql_select(table, columns)
    data = read_civis_sql(
        sql=sql,
        database=database,
        return_as=return_as,
        encoding=encoding,
        job_name=job_name,
        client=client,
        credential_id=credential_id,
        polling_interval=polling_interval,
        hidden=hidden,
        **kwargs,
    )
    return data


def export_to_civis_file(
    sql: str,
    database: str | int,
    sql_params_arguments: dict | None = None,
    job_name: str | None = None,
    client: APIClient | None = None,
    credential_id: int | None = None,
    polling_interval: int | float | None = None,
    hidden: bool = True,
    csv_settings: dict | None = None,
) -> CivisFuture:
    """Store results of a query to a Civis file

    Parameters
    ----------
    sql : str
        The SQL select string to be executed.
    database : str or int
        Execute the query against this database. Can be the database name
        or ID.
    sql_params_arguments : dict, optional
        A dictionary of SQL query parameters to pass directly to
        :func:`civis.APIClient.scripts.post_sql<civis.resources._resources.Scripts.post_sql>`.
        The only allowed keys are ``"params"`` (whose value is a list[dict]) and
        ``"arguments"`` (whose value is a dict). Please refer to the linked API
        documentation for how to format these two keys' values.
    job_name : str, optional
        A name to give the job. If omitted, a random job name will be
        used.
    client : :class:`civis.APIClient`, optional
        If not provided, an :class:`civis.APIClient` object will be
        created from the :envvar:`CIVIS_API_KEY`.
    credential_id : str or int, optional
        The database credential ID.  If ``None``, the default credential
        will be used.
    polling_interval : int or float, optional
        Number of seconds to wait between checks for query completion.
    hidden : bool, optional
        If ``True`` (the default), this job will not appear in the Civis UI.
    csv_settings : dict, optional
        A dictionary of csv_settings to pass to
        :func:`civis.APIClient.scripts.post_sql`.

    Returns
    -------
    fut : :class:`~civis.futures.CivisFuture`
        A future which returns the response from
        :func:`civis.APIClient.scripts.get_sql_runs` after the sql query
        has completed and the result has been stored as a Civis file.

    Examples
    --------
    >>> import civis
    >>> sql = "SELECT * FROM schema.table"
    >>> fut = civis.io.export_to_civis_file(sql, "my_database")
    >>> file_id = fut.result()['output'][0]["file_id"]


    See Also
    --------
    civis.io.read_civis : Read directly into memory without SQL.
    civis.io.read_civis_sql : Read results of a SQL query into memory.
    civis.io.civis_to_csv : Write directly to a CSV file.
    civis.io.civis_file_to_table : Upload a Civis file to a Civis table
    """
    client = client or APIClient()
    script_id, run_id = _sql_script(
        client=client,
        sql=sql,
        database=database,
        job_name=job_name,
        credential_id=credential_id,
        csv_settings=csv_settings,
        hidden=hidden,
        sql_params_arguments=sql_params_arguments,
    )
    fut = CivisFuture(
        client.scripts.get_sql_runs,
        (script_id, run_id),
        polling_interval=polling_interval,
        client=client,
        poll_on_creation=False,
    )
    return fut


def read_civis_sql(
    sql: str,
    database: str | int,
    return_as: str = "list",
    use_pandas=DeprecatedKwargDefault(),
    sql_params_arguments: dict | None = None,
    encoding: str | None = None,
    job_name: str | None = None,
    client: APIClient | None = None,
    credential_id: int | None = None,
    polling_interval: int | float | None = None,
    hidden: bool = True,
    **kwargs,
) -> list | pd.DataFrame | pl.DataFrame:
    """Read data from Civis using a custom SQL string.

    If no data is expected to return from the query,
    consider :func:`~civis.io.query_civis` instead.

    Parameters
    ----------
    sql : str
        The SQL select string to be executed.
    database : str or int
        Execute the query against this database. Can be the database name
        or ID.
    return_as : str, {"list", "pandas", "polars"}
        If ``"list"`` (the default), return a list.
        If ``"pandas"``, return a :class:`pandas.DataFrame`.
        If ``"polars"``, return a :class:`polars.DataFrame`.
    use_pandas : bool, optional
        If ``True``, return a :class:`pandas:pandas.DataFrame`. Otherwise,
        return a list of results from :func:`python:csv.reader`.

        .. deprecated:: 2.6.0
            ``use_pandas`` will be removed at civis-python v3.0.0.
            Please use ``return_as`` instead.
    sql_params_arguments : dict, optional
        A dictionary of SQL query parameters to pass directly to
        :func:`civis.APIClient.scripts.post_sql<civis.resources._resources.Scripts.post_sql>`.
        The only allowed keys are ``"params"`` (whose value is a list[dict]) and
        ``"arguments"`` (whose value is a dict). Please refer to the linked API
        documentation for how to format these two keys' values.
    encoding : str, optional
        If ``return_as`` is ``"pandas"`` or ``"polars"``, this parameter is passed to
        the ``encoding`` kwarg of :func:`pandas.read_csv` or
        :func:`polars.read_csv`, respectively.
        If ``return_as`` is ``"list"``, and if this parameter isn't provided,
        then the UTF-8 encoding is assumed. In case you encounter
        a ``UnicodeDecodeError``, consider choosing an encoding suitable
        for your data; see the `list of standard encodings
        <https://docs.python.org/3/library/codecs.html#standard-encodings>`_.
    job_name : str, optional
        A name to give the job. If omitted, a random job name will be
        used.
    client : :class:`civis.APIClient`, optional
        If not provided, an :class:`civis.APIClient` object will be
        created from the :envvar:`CIVIS_API_KEY`.
    credential_id : str or int, optional
        The database credential ID.  If ``None``, the default credential
        will be used.
    polling_interval : int or float, optional
        Number of seconds to wait between checks for query completion.
    hidden : bool, optional
        If ``True`` (the default), this job will not appear in the Civis UI.
    **kwargs : kwargs
        Extra keyword arguments are passed into
        :func:`pandas.read_csv` if ``return_as`` is ``"pandas"``, or
        passed into :func:`python:csv.reader` if ``return_as`` is ``"list"``,
        or passed into :func:`polars.read_csv` if ``return_as`` is ``"polars"``.

    Returns
    -------
    data : list | :class:`pandas.DataFrame` | :class:`polars.DataFrame`
        A list of rows (with header as first row) if ``return_as`` is
        ``"list"``, or :class:`pandas.DataFrame` if ``return_as`` is ``"pandas"``,
        or :class:`polars.DataFrame` if ``return_as`` is ``"polars"``.
        Note that if ``return_as`` is ``"list"``, no parsing of types is performed and
        each row will be a list of strings.

    Raises
    ------
    ImportError
        If ``return_as`` is ``"pandas"`` and pandas is not installed.
        If ``return_as`` is ``"polars"`` and polars is not installed.
    EmptyResultError
        If no rows were returned as a result of the query.

    Examples
    --------
    >>> import civis
    >>> sql = "SELECT * FROM schema.table"
    >>> df = civis.io.read_civis_sql(sql, "my_database", return_as="pandas")
    >>> col_a = df["column_a"]

    >>> data = civis.io.read_civis_sql(sql, "my_database")
    >>> columns = data.pop(0)
    >>> col_a_index = columns.index("column_a")
    >>> col_a = [row[col_a_index] for row in data]

    Notes
    -----
    This reads the data into memory.

    See Also
    --------
    civis.io.read_civis : Read directly into memory without SQL.
    civis.io.civis_to_csv : Write directly to a CSV file.
    """
    _validate_return_as(return_as)
    return_as = _warn_or_raise_for_use_pandas(use_pandas, return_as)

    if client is None:
        client = APIClient()

    db_id = client.get_database_id(database)
    credential_id = credential_id or client.default_database_credential_id

    script_id, run_id = _sql_script(
        client,
        sql,
        db_id,
        job_name,
        credential_id,
        csv_settings={"compression": "gzip"},
        hidden=hidden,
        sql_params_arguments=sql_params_arguments,
    )
    fut = CivisFuture(
        client.scripts.get_sql_runs,
        (script_id, run_id),
        polling_interval=polling_interval,
        client=client,
        poll_on_creation=False,
    )
    fut.result()
    outputs = client.scripts.get_sql_runs(script_id, run_id)["output"]
    if not outputs:
        raise EmptyResultError("Query {} returned no output.".format(script_id))

    url = outputs[0]["path"]
    file_id = outputs[0]["file_id"]
    log.debug(
        "Exported results to Civis file %s (%s)", outputs[0]["output_name"], file_id
    )

    if return_as == "pandas":
        kwargs["compression"] = "gzip"
        kwargs["encoding"] = encoding

        data = pd.read_csv(url, **kwargs)
    elif return_as == "polars":
        kwargs["encoding"] = encoding
        data = pl.read_csv(url, **kwargs)
    else:
        for attempt in get_default_retrying():
            with attempt:
                response = requests.get(url, stream=True, timeout=60)
                response.raise_for_status()

        with io.StringIO() as buf:
            _decompress_stream(
                response, buf, write_bytes=False, encoding=encoding or "utf-8"
            )
            buf.seek(0)
            data = list(csv.reader(buf, **kwargs))

    return data


def civis_to_csv(
    filename: str,
    sql: str,
    database: str | int,
    sql_params_arguments: dict | None = None,
    job_name: str | None = None,
    client: APIClient | None = None,
    credential_id: int | None = None,
    include_header: bool = True,
    compression: str = "none",
    delimiter: str = ",",
    unquoted: bool = False,
    hidden: bool = True,
    polling_interval: int | float | None = None,
) -> CivisFuture:
    """Export data from Civis to a local CSV file.

    Parameters
    ----------
    filename : str
        Download exported data into this file.
    sql : str
        The SQL select string to be executed.
    database : str or int
        Export data from this database. Can be the database name or ID.
    sql_params_arguments : dict, optional
        A dictionary of SQL query parameters to pass directly to
        :func:`civis.APIClient.scripts.post_sql<civis.resources._resources.Scripts.post_sql>`.
        The only allowed keys are ``"params"`` (whose value is a list[dict]) and
        ``"arguments"`` (whose value is a dict). Please refer to the linked API
        documentation for how to format these two keys' values.
    job_name : str, optional
        A name to give the job. If omitted, a random job name will be
        used.
    client : :class:`civis.APIClient`, optional
        If not provided, an :class:`civis.APIClient` object will be
        created from the :envvar:`CIVIS_API_KEY`.
    credential_id : str or int, optional
        The ID of the database credential.  If ``None``, the default
        credential will be used.
    include_header: bool, optional
        If ``True``, the first line of the CSV will be headers.
        Default: ``True``.
    compression: str, optional
        Type of compression to use, if any. One of ``'none'``, ``'zip'``, or
        ``'gzip'``. Default ``'none'``. ``'gzip'`` currently returns a file
        with no compression unless include_header is set to False. In a
        future release, a ``'gzip'`` compressed file will be returned for
        all cases.
    delimiter: str, optional
        Which delimiter to use, if any. One of ``','``, ``'\t'``, or
        ``'|'``. Default: ``','``.
    unquoted: bool, optional
        Whether or not to quote fields. Default: ``False``.
    polling_interval : int or float, optional
        Number of seconds to wait between checks for query completion.
    hidden : bool, optional
        If ``True`` (the default), this job will not appear in the Civis UI.

    Returns
    -------
    results : :class:`~civis.futures.CivisFuture`
        A ``CivisFuture`` object that represents the SQL query that unloads data
        from Civis Platform. Note that this unloading process has an arbitrary
        filename assigned and made available through the ``CivisFuture`` object,
        which is not the same as the `filename` parameter. It is recommended
        that a separate variable is used to store the desired filename beforehand
        so that it's accessible after this function finishes,
        see the code example below.

    Examples
    --------
    >>> import civis
    >>> sql = "SELECT * FROM schema.table"
    >>> file_path = "file.csv"
    >>> fut = civis.io.civis_to_csv(file_path, sql, "my_database")
    >>> fut.result()  # Wait for job to complete
    >>> import pandas as pd  # Let's say we want to read the data into a DataFrame
    >>> df = pd.read_csv(file_path)

    See Also
    --------
    civis.io.read_civis : Read table contents into memory.
    civis.io.read_civis_sql : Read results of a SQL query into memory.
    civis.io.export_to_civis_file : Store a SQL query's results in a Civis file
    """
    if client is None:
        client = APIClient()

    db_id = client.get_database_id(database)
    credential_id = credential_id or client.default_database_credential_id

    # don't fix bug that would cause breaking change for now
    # when gzip compression is requested, a gzip file is not actually returned
    # instead the gzip file is decompressed during download
    if compression == "gzip" and include_header:
        compression = "none"

    # don't support parallel unload; the output format
    # is different which would introduce a breaking change
    headers = b""

    try:
        delimiter = DELIMITERS[delimiter]
    except KeyError:
        raise ValueError("delimiter must be one of {}".format(DELIMITERS.keys()))

    # always set compression to gzip to reduce I/O
    csv_settings = dict(
        include_header=include_header,
        compression="gzip",
        column_delimiter=delimiter,
        unquoted=unquoted,
        filename_prefix=None,
        force_multifile=False,
    )

    script_id, run_id = _sql_script(
        client,
        sql,
        db_id,
        job_name,
        credential_id,
        hidden=hidden,
        csv_settings=csv_settings,
        sql_params_arguments=sql_params_arguments,
    )
    fut = CivisFuture(
        client.scripts.get_sql_runs,
        (script_id, run_id),
        polling_interval=polling_interval,
        client=client,
        poll_on_creation=False,
    )
    download = _download_callback(script_id, run_id, filename, headers, compression)
    fut.add_done_callback(download)

    return fut


def civis_to_multifile_csv(
    sql: str,
    database: str | int,
    sql_params_arguments: dict | None = None,
    job_name: str | None = None,
    client: APIClient | None = None,
    credential_id: int | None = None,
    include_header: bool = True,
    compression: str = "none",
    delimiter: str = "|",
    max_file_size: int | None = None,
    unquoted: bool = False,
    prefix: str | None = None,
    polling_interval: int | float | None = None,
    hidden: bool = True,
) -> dict:
    """Unload the result of SQL query and return presigned urls.

    This function is intended for unloading large queries/tables from redshift
    as it uses a 'PARALLEL ON' S3 unload. It returns a similar manifest file
    to conventional S3 UNLOAD statements except the CSV parts are accessible
    via both files endpoint IDs and presigned S3 urls.

    Parameters
    ----------
    sql : str
        The SQL select string to be executed.
    database : str or int
        Execute the query against this database. Can be the database name
        or ID.
    sql_params_arguments : dict, optional
        A dictionary of SQL query parameters to pass directly to
        :func:`civis.APIClient.scripts.post_sql<civis.resources._resources.Scripts.post_sql>`.
        The only allowed keys are ``"params"`` (whose value is a list[dict]) and
        ``"arguments"`` (whose value is a dict). Please refer to the linked API
        documentation for how to format these two keys' values.
    job_name : str, optional
        A name to give the job. If omitted, a random job name will be
        used.
    client : :class:`civis.APIClient`, optional
        If not provided, an :class:`civis.APIClient` object will be
        created from the :envvar:`CIVIS_API_KEY`.
    credential_id : str or int, optional
        The database credential ID.  If ``None``, the default credential
        will be used.
    include_header: bool, optional
        If ``True`` include a key in the returned dictionary containing a list
        of column names. Default: ``True``.
    compression: str, optional
        Type of compression to use, if any. One of ``'none'``, ``'zip'``, or
        ``'gzip'``. Default ``'none'``.
    delimiter: str, optional
        Which delimiter to use, if any. One of ``','``, ``'\t'``, or
        ``'|'``. Default: ``'|'``.
    max_file_size: int, optional
        Maximum number of Megabytes each created file will be.
    unquoted: bool, optional
        Whether or not to quote fields. Default: ``False``.
    prefix: str, optional
        A user specified filename prefix for the output file to have. Default:
        ``None``.
    polling_interval : int or float, optional
        Number of seconds to wait between checks for query completion.
    hidden : bool, optional
        If ``True`` (the default), this job will not appear in the Civis UI.

    Returns
    -------
    unload_manifest: dict
        A dictionary resembling an AWS manifest file. Has the following keys:

        'query': str
            The query.

        'header': list of str
            The columns from the query.

        'entries': list of dict
            Each dict has the following keys:

            'id': int
                File ID
            'name': str
                Filename
            'size': int
                File size in bytes
            'url': str
                Unsigned S3 URL ('s3://...')
            'url_signed': str
                Signed S3 URL ('https://...')

        'unquoted': bool
            Whether the cells are quoted.

        'compression': str
            Type of compression used.

        'delimiter': str
            Delimiter that separates the cells.

    Examples
    --------
    >>> import io
    >>> import civis
    >>> import pandas as pd
    >>> sql = "SELECT * FROM schema.my_big_table"
    >>> database = "my_database"
    >>> delimiter = "|"
    >>> manifest = civis.io.civis_to_multifile_csv(sql, database, delimiter=delimiter)
    >>> ids = [entry['id'] for entry in manifest['entries']]
    >>> for file_id in ids:
    >>>     buf = io.BytesIO()
    >>>     civis.io.civis_to_file(file_id, buf)
    >>>     buf.seek(0)
    >>>     # Process the data in `buf` for your own application, e.g.:
    >>>     df = pd.read_csv(buf, delimiter=delimiter)

    See Also
    --------
    civis.APIClient.scripts.post_sql
    """
    if client is None:
        client = APIClient()
    try:
        delimiter = DELIMITERS[delimiter]
    except KeyError:
        raise ValueError(f"delimiter must be one of {DELIMITERS.keys()}: {delimiter}")

    csv_settings = dict(
        include_header=include_header,
        compression=compression,
        column_delimiter=delimiter,
        unquoted=unquoted,
        filename_prefix=prefix,
        force_multifile=True,
        max_file_size=max_file_size,
    )
    script_id, run_id = _sql_script(
        client,
        sql,
        database,
        job_name,
        credential_id,
        hidden,
        csv_settings=csv_settings,
        sql_params_arguments=sql_params_arguments,
    )

    fut = CivisFuture(
        client.scripts.get_sql_runs,
        (script_id, run_id),
        polling_interval=polling_interval,
        client=client,
        poll_on_creation=False,
    )

    outputs = fut.result()["output"]
    if not outputs:
        raise EmptyResultError(
            "Unload query {} returned no manifest.".format(script_id)
        )

    buf = io.BytesIO()
    civis_to_file(outputs[0]["file_id"], buf, client=client)
    txt = io.TextIOWrapper(buf, encoding="utf-8")
    txt.seek(0)
    unload_manifest = json.load(txt)

    return unload_manifest


def dataframe_to_civis(
    df: pd.DataFrame | pl.DataFrame,
    database: str | int,
    table: str,
    client: APIClient | None = None,
    max_errors: int | None = None,
    existing_table_rows: str = "fail",
    diststyle: str | None = None,
    distkey: str | None = None,
    sortkey1: str | None = None,
    sortkey2: str | None = None,
    table_columns: list[dict[str, str]] | None = None,
    credential_id: int | None = None,
    primary_keys: list[str] | None = None,
    last_modified_keys: list[str] | None = None,
    execution: str = "immediate",
    polling_interval: int | float | None = None,
    hidden: bool = True,
    **kwargs,
) -> CivisFuture:
    """Upload a dataframe into a Civis table.

    For a pandas dataframe,
    the dataframe's index will not be included. To store the index
    along with the other values, use ``df.reset_index()`` instead
    of ``df`` as the first argument to this function.

    Parameters
    ----------
    df : :class:`pandas.DataFrame` | :class:`polars.DataFrame`
        The dataframe to upload to Civis.
    database : str or int
        Upload data into this database. Can be the database name or ID.
    table : str
        The schema and table you want to upload to. E.g.,
        ``'scratch.table'``. Schemas or tablenames with periods must
        be double quoted, e.g. ``'scratch."my.table"'``.
    client : :class:`civis.APIClient`, optional
        If not provided, an :class:`civis.APIClient` object will be
        created from the :envvar:`CIVIS_API_KEY`.
    max_errors : int, optional
        The maximum number of rows with errors to remove from the import
        before failing.
    existing_table_rows : str, optional
        The behaviour if a table with the requested name already exists.
        One of ``'fail'``, ``'truncate'``, ``'append'``, ``'drop'``, or
        ``'upsert'``. Defaults to ``'fail'``.
    diststyle : str, optional
        The distribution style for the table.
        One of ``'even'``, ``'all'`` or ``'key'``.
    distkey : str, optional
        The column to use as the distkey for the table.
    sortkey1 : str, optional
        The column to use as the sortkey for the table.
    sortkey2 : str, optional
        The second column in a compound sortkey for the table.
    table_columns : list[Dict[str, str]], optional
        A list of dictionaries, ordered so each dictionary corresponds
        to a column in the order that it appears in the source file. Each dict
        should have a key "name" that corresponds to the column name in the
        destination table, and a key "sql_type" corresponding to the intended
        column data type in the destination table. The "sql_type" key is not
        required when appending to an existing table. The table_columns
        parameter is required if the table does not exist, the table is being
        dropped, or the columns in the source file do not appear in the same
        order as in the destination table. Example:
        ``[{"name": "foo", "sql_type": "INT"}, {"name": "bar", "sql_type": "VARCHAR"}]``
    credential_id : str or int, optional
        The ID of the database credential.  If ``None``, the default
        credential will be used.
    primary_keys: list[str], optional
        A list of the primary key column(s) of the destination table that
        uniquely identify a record. These columns must not contain null values.
        If existing_table_rows is "upsert", this
        field is required. Note that this is true regardless of whether the
        destination database itself requires a primary key.
    last_modified_keys: list[str], optional
        A list of the columns indicating a record has been updated. If
        existing_table_rows is "upsert", this field is required.
    execution: string, optional, default "immediate"
        One of "delayed" or "immediate". If "immediate", refresh column
        statistics as part of the run. If "delayed", flag the table for a
        deferred statistics update; column statistics may not be available
        for up to 24 hours. In addition, if existing_table_rows is "upsert",
        delayed executions move data from staging table to final table after a
        brief delay, in order to accommodate multiple concurrent imports to the
        same destination table.
    polling_interval : int or float, optional
        Number of seconds to wait between checks for job completion.
    hidden : bool, optional
        If ``True`` (the default), this job will not appear in the Civis UI.
    **kwargs : kwargs
        Extra keyword arguments will be passed to
        :func:`pandas.DataFrame.to_csv`
        or :func:`polars.DataFrame.write_csv`.

    Returns
    -------
    fut : :class:`~civis.futures.CivisFuture`
        A ``CivisFuture`` object.

    Examples
    --------
    >>> import civis
    >>> import pandas as pd
    >>> df = pd.DataFrame({'a': [1, 2, 3], 'b': [4, 5, 6]})
    >>> fut = civis.io.dataframe_to_civis(df, 'my-database',
    ...                                   'scratch.df_table')
    >>> fut.result()

    See Also
    --------
    :func:`pandas.DataFrame.to_csv`
    :func:`polars.DataFrame.write_csv`
    """  # noqa: E501
    if client is None:
        client = APIClient()

    headers = False if kwargs.get("header") is False else True
    with TemporaryDirectory() as tmp_dir:
        tmp_path = os.path.join(tmp_dir, "dataframe_to_civis.csv")
        if HAS_PANDAS and isinstance(df, pd.DataFrame):
            to_csv_kwargs = {"encoding": "utf-8", "index": False}
            to_csv_kwargs.update(kwargs)
            df.to_csv(tmp_path, **to_csv_kwargs)  # type: ignore[call-overload]
        elif HAS_POLARS and isinstance(df, pl.DataFrame):
            df.write_csv(tmp_path, **kwargs)
        else:
            raise ValueError(
                f"unsuppported dataframe type {type(df)!r} "
                "-- only pandas and polars are supported"
            )
        _, name = split_schema_tablename(table)
        file_id = file_to_civis(tmp_path, name, client=client)

    delimiter = ","
    fut = civis_file_to_table(
        file_id,
        database,
        table,
        client=client,
        max_errors=max_errors,
        existing_table_rows=existing_table_rows,
        diststyle=diststyle,
        distkey=distkey,
        sortkey1=sortkey1,
        sortkey2=sortkey2,
        table_columns=table_columns,
        delimiter=delimiter,
        headers=headers,
        credential_id=credential_id,
        primary_keys=primary_keys,
        last_modified_keys=last_modified_keys,
        escaped=False,
        execution=execution,
        polling_interval=polling_interval,
        hidden=hidden,
    )

    return fut


def csv_to_civis(
    filename: str,
    database: str | int,
    table: str,
    client: APIClient | None = None,
    max_errors: int | None = None,
    existing_table_rows: str = "fail",
    diststyle: str | None = None,
    distkey: str | None = None,
    sortkey1: str | None = None,
    sortkey2: str | None = None,
    table_columns: list[dict] | None = None,
    delimiter: str = ",",
    headers: bool | None = None,
    primary_keys: list[str] | None = None,
    last_modified_keys: list[str] | None = None,
    escaped: bool = False,
    execution: str = "immediate",
    credential_id: int | None = None,
    polling_interval: int | float | None = None,
    hidden: bool = True,
) -> CivisFuture:
    """Upload the contents of a local CSV file to Civis.

    Parameters
    ----------
    filename : str
        Upload the contents of this file.
    database : str or int
        Upload data into this database. Can be the database name or ID.
    table : str
        The schema and table you want to upload to. E.g.,
        ``'scratch.table'``.
    client : :class:`civis.APIClient`, optional
        If not provided, an :class:`civis.APIClient` object will be
        created from the :envvar:`CIVIS_API_KEY`.
    max_errors : int, optional
        The maximum number of rows with errors to remove from the import
        before failing.
    existing_table_rows : str, optional
        The behaviour if a table with the requested name already exists.
        One of ``'fail'``, ``'truncate'``, ``'append'``, ``'drop'``, or
        ``'upsert'``. Defaults to ``'fail'``.
    diststyle : str, optional
        The distribution style for the table.
        One of ``'even'``, ``'all'`` or ``'key'``.
    distkey : str, optional
        The column to use as the distkey for the table.
    sortkey1 : str, optional
        The column to use as the sortkey for the table.
    sortkey2 : str, optional
        The second column in a compound sortkey for the table.
    table_columns : list[Dict[str, str]], optional
        A list of dictionaries, ordered so each dictionary corresponds
        to a column in the order that it appears in the source file. Each dict
        should have a key "name" that corresponds to the column name in the
        destination table, and a key "sql_type" corresponding to the intended
        column data type in the destination table. The "sql_type" key is not
        required when appending to an existing table. The table_columns
        parameter is required if the table does not exist, the table is being
        dropped, or the columns in the source file do not appear in the same
        order as in the destination table. Example:
        ``[{"name": "foo", "sql_type": "INT"}, {"name": "bar", "sql_type": "VARCHAR"}]``
    delimiter : string, optional
        The column delimiter. One of ``','``, ``'\\t'`` or ``'|'``.
    headers : bool, optional
        Whether or not the first row of the file should be treated as
        headers. The default, ``None``, attempts to autodetect whether
        or not the first row contains headers.
    primary_keys: list[str], optional
        A list of the primary key column(s) of the destination table that
        uniquely identify a record. These columns must not contain null values.
        If existing_table_rows is "upsert", this
        field is required. Note that this is true regardless of whether the
        destination database itself requires a primary key.
    last_modified_keys: list[str], optional
        A list of the columns indicating a record has been updated. If
        existing_table_rows is "upsert", this field is required.
    escaped: bool, optional
        A boolean value indicating whether or not the source file has quotes
        escaped with a backslash. Defaults to false.
    execution: string, optional, default "immediate"
        One of "delayed" or "immediate". If "immediate", refresh column
        statistics as part of the run. If "delayed", flag the table for a
        deferred statistics update; column statistics may not be available
        for up to 24 hours. In addition, if existing_table_rows is "upsert",
        delayed executions move data from staging table to final table after a
        brief delay, in order to accommodate multiple concurrent imports to the
        same destination table.
    credential_id : str or int, optional
        The ID of the database credential.  If ``None``, the default
        credential will be used.
    polling_interval : int or float, optional
        Number of seconds to wait between checks for job completion.
    hidden : bool, optional
        If ``True`` (the default), this job will not appear in the Civis UI.

    Returns
    -------
    results : :class:`~civis.futures.CivisFuture`
        A `CivisFuture` object.

    Notes
    -----
    This reads the contents of `filename` into memory.

    Examples
    --------
    >>> import civis
    >>> with open('input_file.csv', 'w') as _input:
    ...     _input.write('a,b,c\\n1,2,3')
    >>> fut = civis.io.csv_to_civis('input_file.csv',
    ...                             'my-database',
    ...                             'scratch.my_data')
    >>> fut.result()
    """  # noqa: E501
    if client is None:
        client = APIClient()

    name = path.basename(filename)
    with open(filename, "rb") as data:
        file_id = file_to_civis(data, name, client=client)
        log.debug("Uploaded file %s to Civis file %s", filename, file_id)
        fut = civis_file_to_table(
            file_id,
            database,
            table,
            client=client,
            max_errors=max_errors,
            existing_table_rows=existing_table_rows,
            diststyle=diststyle,
            distkey=distkey,
            sortkey1=sortkey1,
            sortkey2=sortkey2,
            table_columns=table_columns,
            delimiter=delimiter,
            headers=headers,
            credential_id=credential_id,
            primary_keys=primary_keys,
            last_modified_keys=last_modified_keys,
            escaped=escaped,
            execution=execution,
            polling_interval=polling_interval,
            hidden=hidden,
        )
    return fut


def civis_file_to_table(
    file_id: int | list[int],
    database: str | int,
    table: str,
    client: APIClient | None = None,
    max_errors: int | None = None,
    existing_table_rows: str = "fail",
    diststyle: str | None = None,
    distkey: str | None = None,
    sortkey1: str | None = None,
    sortkey2: str | None = None,
    table_columns: list[dict[str, str]] | None = None,
    primary_keys: list[str] | None = None,
    last_modified_keys: list[str] | None = None,
    escaped: bool = False,
    execution: str = "immediate",
    delimiter: str | None = None,
    headers: bool | None = None,
    credential_id: int | None = None,
    polling_interval: int | float | None = None,
    hidden: bool = True,
) -> CivisFuture:
    """Upload the contents of one or more Civis files to a Civis table.
    All provided files will be loaded as an atomic unit in parallel, and
    should share the same columns in the same order, and be in the same
    format.

    .. note::
        The input Civis files must be in a CSV-like delimiter separated format and
        will be accepted in both uncompressed and compressed format
        (.zip, .gz).

    Parameters
    ----------
    file_id : int or list[int]
        Civis file ID or a list of Civis file IDs.
    database : str or int
        Upload data into this database. Can be the database name or ID.
    table : str
        The schema and table you want to upload to. E.g.,
        ``'scratch.table'``.
    client : :class:`civis.APIClient`, optional
        If not provided, an :class:`civis.APIClient` object will be
        created from the :envvar:`CIVIS_API_KEY`.
    max_errors : int, optional
        The maximum number of rows with errors to remove from the import
        before failing. If multiple files are provided, this limit applies
        across all files combined.
    existing_table_rows : str, optional
        The behaviour if a table with the requested name already exists.
        One of ``'fail'``, ``'truncate'``, ``'append'``, ``'drop'``, or
        ``'upsert'``. Defaults to ``'fail'``.
    diststyle : str, optional
        The distribution style for the table.
        One of ``'even'``, ``'all'`` or ``'key'``.
    distkey : str, optional
        The column to use as the distkey for the table.
    sortkey1 : str, optional
        The column to use as the sortkey for the table.
    sortkey2 : str, optional
        The second column in a compound sortkey for the table.
    table_columns : list[Dict[str, str]], optional
        A list of dictionaries, ordered so each dictionary corresponds
        to a column in the order that it appears in the source file. Each dict
        should have a key "name" that corresponds to the column name in the
        destination table, and a key "sql_type" corresponding to the intended
        column data type in the destination table. The "sql_type" key is not
        required when appending to an existing table. The table_columns
        parameter is required if the table does not exist, the table is being
        dropped, or the columns in the source file do not appear in the same
        order as in the destination table. Example:
        ``[{"name": "foo", "sql_type": "INT"}, {"name": "bar", "sql_type": "VARCHAR"}]``
    primary_keys: list[str], optional
        A list of the primary key column(s) of the destination table that
        uniquely identify a record. These columns must not contain null values.
        If existing_table_rows is "upsert", this
        field is required. Note that this is true regardless of whether the
        destination database itself requires a primary key.
    last_modified_keys: list[str], optional
        A list of the columns indicating a record has been updated. If
        existing_table_rows is "upsert", this field is required.
    escaped: bool, optional
        A boolean value indicating whether or not the source file(s) escape
        quotes with a backslash. Defaults to false.
    execution: string, optional, default "immediate"
        One of "delayed" or "immediate". If "immediate", refresh column
        statistics as part of the run. If "delayed", flag the table for a
        deferred statistics update; column statistics may not be available
        for up to 24 hours. In addition, if existing_table_rows is "upsert",
        delayed executions move data from staging table to final table after a
        brief delay, in order to accommodate multiple concurrent imports to the
        same destination table.
    delimiter : string, optional
        The column delimiter. One of ``','``, ``'\\t'`` or ``'|'``. If not
        provided, will attempt to auto-detect.
    headers : bool, optional
        Whether or not the first row of the file should be treated as
        headers. The default, ``None``, attempts to autodetect whether
        or not the first row contains headers.
    credential_id : str or int, optional
        The ID of the database credential.  If ``None``, the default
        credential will be used.
    polling_interval : int or float, optional
        Number of seconds to wait between checks for job completion.
    hidden : bool, optional
        If ``True`` (the default), this job will not appear in the Civis UI.

    Returns
    -------
    results : :class:`~civis.futures.CivisFuture`
        A `CivisFuture` object.

    Raises
    ------
    CivisImportError
        If multiple files are given and determined to be incompatible for
        import. This may be the case if their columns have different types,
        their delimiters are different, headers are present in some but not
        others, or compressions do not match.
    TypeError
        If the type of the file_id parameter is a string. This situation may
        arise when the file ID comes from an environment variable and is not
        cast from string to integer before being passed to civis_file_to_table.

    Examples
    --------
    >>> import civis
    >>> file_id = 100
    >>> fut = civis.io.civis_file_to_table(file_id,
    ...                                    'my-database',
    ...                                    'scratch.my_data')
    >>> fut.result()
    """  # noqa: E501
    if client is None:
        client = APIClient()

    if type(file_id) is str:
        raise TypeError("Invalid type for file_id: str. " "Must be int or list[int]")

    schema, table_name = split_schema_tablename(table)
    if isinstance(file_id, int):
        file_id = [file_id]
    if not file_id:
        raise ValueError("Provide one or multiple meaningful input file IDs.")
    if schema is None:
        raise ValueError("Provide a schema as part of the `table` input.")
    db_id = client.get_database_id(database)
    cred_id = credential_id or client.default_database_credential_id
    if delimiter is not None:  # i.e. it was provided as an argument
        delimiter = DELIMITERS.get(delimiter)
        if not delimiter:
            raise ValueError(
                f"delimiter must be one of {DELIMITERS.keys()}: {delimiter}"
            )
    if table_columns:
        # If the data cleaning code doesn't find a "sql_type" for each
        # entry, it will silently replace the input table_columns with
        # an inferred table_columns. Make sure there's no typos in the input.
        keys = set(key for hash in table_columns for key in hash)
        valid_keys = {"name", "sql_type"}
        invalid_keys = keys - valid_keys
        if invalid_keys:
            # Sort the sets for display to allow for deterministic testing in
            # Python versions < 3.7.
            raise ValueError(
                "Keys of the dictionaries contained in `table_columns` must "
                "be one of {}. The input `table_columns` also has "
                "{}.".format(tuple(sorted(valid_keys)), tuple(sorted(invalid_keys)))
            )

    try:
        client.databases.get_schemas_tables(db_id, schema, table_name)
        log.debug(
            "Table {table} already exists - skipping column "
            "detection".format(table=table)
        )
        table_exists = True
    except CivisAPIError as e:
        table_exists = False
        if e.status_code != 404:
            warnings.warn(
                "Unexpected error when checking if table %s.%s "
                "exists on database %d:\n%s" % (schema, table_name, db_id, str(e))
            )

    sql_types_provided = False
    if table_columns:
        sql_type_cnt = sum(1 for col in table_columns if col.get("sql_type"))
        if sql_type_cnt == len(table_columns):
            sql_types_provided = True
        elif sql_type_cnt != 0:
            error_message = (
                "Some table columns " "have a sql type provided, " "but others do not."
            )
            raise ValueError(error_message)

    # Use Preprocess endpoint to get the table columns as needed
    # and perform necessary file cleaning
    need_table_columns = (not table_exists or existing_table_rows == "drop") and (
        not sql_types_provided
    )

    cleaning_futures = _run_cleaning(
        file_id, client, need_table_columns, headers, delimiter, hidden
    )

    (cleaned_file_ids, headers, compression, delimiter, cleaned_table_columns) = (
        _process_cleaning_results(
            cleaning_futures, client, headers, need_table_columns, delimiter
        )
    )

    table_columns = cleaned_table_columns if need_table_columns else table_columns

    source = dict(file_ids=cleaned_file_ids)
    destination = dict(
        schema=schema,
        table=table_name,
        remote_host_id=db_id,
        credential_id=cred_id,
        primary_keys=primary_keys,
        last_modified_keys=last_modified_keys,
    )

    redshift_options = dict(
        distkey=distkey, sortkeys=[sortkey1, sortkey2], diststyle=diststyle
    )

    import_name = "CSV import to {}.{}".format(schema, table_name)
    import_job = client.imports.post_files_csv(
        source,
        destination,
        headers,
        name=import_name,
        max_errors=max_errors,
        existing_table_rows=existing_table_rows,
        column_delimiter=delimiter,
        compression=compression,
        escaped=escaped,
        execution=execution,
        # If the user hasn't explicitly provided table column info,
        # then there might be differences in their precisions/lengths.
        # Setting this option will allow the Civis API
        # to increase these values for the data types provided,
        # and decreases the risk of a length-related import failure
        # when types are inferred.
        loosen_types=need_table_columns,
        table_columns=table_columns,
        redshift_destination_options=redshift_options,
        hidden=hidden,
    )
    fut = run_job(import_job.id, client=client, polling_interval=polling_interval)
    log.debug("Started run %d for import %d", fut.run_id, import_job.id)
    return fut


def _sql_script(
    client,
    sql,
    database,
    job_name,
    credential_id,
    hidden=False,
    csv_settings=None,
    sql_params_arguments=None,
):
    job_name = maybe_get_random_name(job_name)
    db_id = client.get_database_id(database)
    credential_id = credential_id or client.default_database_credential_id
    csv_settings = csv_settings or {}
    sql_params_arguments = sql_params_arguments or {}

    if not (keys := set(sql_params_arguments.keys())) <= _SQL_PARAMS_ARGUMENTS_KEYS:
        raise ValueError(
            "sql_params_arguments must be a dict with keys in "
            f"{set(_SQL_PARAMS_ARGUMENTS_KEYS)}: {keys}"
        )

    export_job = client.scripts.post_sql(
        job_name,
        remote_host_id=db_id,
        credential_id=credential_id,
        sql=sql,
        hidden=hidden,
        csv_settings=csv_settings,
        **sql_params_arguments,
    )

    run_job = client.scripts.post_sql_runs(export_job.id)
    log.debug("Started run %d of SQL script %d", run_job.id, export_job.id)
    return export_job.id, run_job.id


def _get_sql_select(table, columns=None):
    if columns and not isinstance(columns, (list, tuple)):
        raise TypeError("columns must be a list, tuple or None")
    select = ", ".join(columns) if columns is not None else "*"
    sql = "select {} from {}".format(select, table)  # nosec
    return sql


def _decompress_stream(response, buf, write_bytes=True, encoding="utf-8"):

    # use response.raw for a more consistent approach
    # if content-encoding is specified in the headers
    # then response.iter_content will decompress the stream
    # however, our use of content-encoding is inconsistent
    chunk = response.raw.read(CHUNK_SIZE)
    d = zlib.decompressobj(zlib.MAX_WBITS | 32)

    while chunk or d.unused_data:
        if d.unused_data:
            to_decompress = d.unused_data + chunk
            d = zlib.decompressobj(zlib.MAX_WBITS | 32)
        else:
            to_decompress = d.unconsumed_tail + chunk
        if write_bytes:
            buf.write(d.decompress(to_decompress))
        else:
            buf.write(d.decompress(to_decompress).decode(encoding))
        chunk = response.raw.read(CHUNK_SIZE)


def _download_file(url, local_path, headers, compression):
    for attempt in get_default_retrying():
        with attempt:
            response = requests.get(url, stream=True, timeout=60)
            response.raise_for_status()

    # gzipped buffers can be concatenated so write headers as gzip
    if compression == "gzip":
        with gzip.open(local_path, "wb") as fout:
            fout.write(headers)
        with open(local_path, "ab") as fout:
            shutil.copyfileobj(response.raw, fout, CHUNK_SIZE)

    # write headers and decompress the stream
    elif compression == "none":
        with open(local_path, "wb") as fout:
            fout.write(headers)
            _decompress_stream(response, fout)

    # decompress the stream, write headers, and zip the file
    elif compression == "zip":
        with TemporaryDirectory() as tmp_dir:
            tmp_path = path.join(tmp_dir, "civis_to_csv.csv")
            with open(tmp_path, "wb") as tmp_file:
                tmp_file.write(headers)
                _decompress_stream(response, tmp_file)

            with zipfile.ZipFile(local_path, "w") as fout:
                arcname = path.basename(local_path)
                if arcname.split(".")[-1] == "zip":
                    arcname = arcname.split(".")[0] + ".csv"
                fout.write(tmp_path, arcname, zipfile.ZIP_DEFLATED)


def _download_callback(job_id, run_id, filename, headers, compression):

    def callback(future):
        if not future.succeeded():
            return
        outputs = future.result().get("output")
        if not outputs:
            warnings.warn(
                "Job %s, run %s does not have any output to "
                "download. Not creating file %s." % (job_id, run_id, filename),
                RuntimeWarning,
            )
            return
        else:
            url = outputs[0]["path"]
            file_id = outputs[0]["file_id"]
            log.debug("Exported results to Civis file %s", file_id)
            return _download_file(url, filename, headers, compression)

    return callback


def split_schema_tablename(table: str) -> tuple[str | None, str]:
    """Split a Redshift 'schema.tablename' string

    Remember that special characters (such as '.') can only
    be included in a schema or table name if delimited by double-quotes.

    Parameters
    ----------
    table: str
        Either a Redshift schema and table name combined
        with a ".", or else a single table name.

    Returns
    -------
    schema, tablename
        A 2-tuple of strings. The ``schema`` may be None if the input
        is only a table name, but the ``tablename`` will always be filled.

    Raises
    ------
    ValueError
        If the input ``table`` is not separable into a schema and
        table name.
    """
    reader = csv.reader(
        io.StringIO(str(table)), delimiter=".", doublequote=True, quotechar='"'
    )
    schema_name_tup = next(reader)
    if len(schema_name_tup) == 1:
        schema_name_tup = (None, schema_name_tup[0])  # type: ignore[assignment]
    if len(schema_name_tup) != 2:
        raise ValueError(
            "Cannot parse schema and table. "
            "Does '{}' follow the pattern 'schema.table'?".format(table)
        )
    return tuple(schema_name_tup)  # type: ignore[return-value]


def _run_cleaning(
    file_ids,
    client,
    need_table_columns,
    headers,
    delimiter,
    hidden,
    polling_interval=None,
):
    cleaning_futures = []
    for fid in file_ids:
        cleaner_job = client.files.post_preprocess_csv(
            file_id=fid,
            in_place=False,
            detect_table_columns=need_table_columns,
            force_character_set_conversion=True,
            include_header=headers,
            column_delimiter=delimiter,
            hidden=hidden,
        )
        fut = run_job(cleaner_job.id, client=client, polling_interval=polling_interval)
        log.debug(
            "Started CSV preprocess job %d run %d for file %d (%s)",
            cleaner_job.id,
            fut.run_id,
            fid,
            client.files.get(fid).name,
        )
        cleaning_futures.append(fut)
    return cleaning_futures


def _process_cleaning_results(
    cleaning_futures, client, headers, need_table_columns, delimiter
):
    futures, _ = concurrent.futures.wait(cleaning_futures)
    files: list[_File] = []

    job_run_ids_no_output = []
    for fut in futures:
        objs = client.jobs.list_runs_outputs(fut.job_id, fut.run_id)
        if not objs:
            job_run_ids_no_output.append((fut.job_id, fut.run_id))
            continue
        # `objs` is guaranteed to have exactly one file output.
        f = client.files.get(objs[0].object_id)
        files.append(_File(id=f.id, name=f.name, detected_info=f.detected_info))
    if job_run_ids_no_output:
        job_run_ids_in_err_msg = "\n".join(
            f"\tjob {j} run {r}" for j, r in job_run_ids_no_output
        )
        raise CivisImportError(
            "No CSV preprocess output found for "
            f"these runs:\n{job_run_ids_in_err_msg}"
        )

    if need_table_columns:
        table_columns, allow_inconsistent_headers = _check_column_types(files)
    else:
        table_columns = None
        allow_inconsistent_headers = False

    try:
        headers = _check_detected_info(files, "includeHeader", headers)
    except CivisImportError:
        if allow_inconsistent_headers:
            headers = True
        else:
            raise

    delimiter = _check_detected_info(files, "columnDelimiter", delimiter)
    compression = _check_detected_info(files, "compression")

    cleaned_file_ids = [f.id for f in files]

    return cleaned_file_ids, headers, compression, delimiter, table_columns


def _check_detected_info(files: list[_File], attr: str, value_from_user=None):
    values_detected = [f.detected_info[attr] for f in files]
    err_msg = _err_msg_if_inconsistent(values_detected, files)
    if err_msg:
        raise CivisImportError(
            f"All detected values for '{attr}' must be the same, "
            f"however --\n{err_msg}"
        )

    value_detected = values_detected[0]

    # The user cannot specify "compression" from civis_file_to_table.
    if attr == "compression" or value_from_user is None:
        return value_detected

    if value_detected != value_from_user:
        raise CivisImportError(
            f"All detected values for '{attr}' in your files "
            f"are {value_detected}, which doesn't match "
            f"your provided value of {value_from_user}"
        )
    else:
        return value_detected


def _check_column_types(files: list[_File]):
    cols_by_file: list[list[dict[str, str]]]
    cols_by_file = [f.detected_info["tableColumns"] for f in files]

    col_counts = [len(cols) for cols in cols_by_file]
    err_msg = _err_msg_if_inconsistent(col_counts, files)
    if err_msg:
        raise CivisImportError(
            f"All files must have the same number of columns, " f"however --\n{err_msg}"
        )

    # Transpose cols_by_file to get cols_by_col
    # https://stackoverflow.com/q/6473679
    cols_by_col: list[list[dict[str, str]]]
    cols_by_col = list(map(list, zip(*cols_by_file)))

    table_columns: list[dict[str, str]] = []
    allow_inconsistent_headers = False
    err_msgs: list[str] = []

    for i, cols in enumerate(cols_by_col, 1):
        col_name = next((name for c in cols if (name := c.get("name"))), f"column_{i}")

        sql_base_types = [col["sql_type"].split("(", 1)[0].upper() for col in cols]
        err_msg = _err_msg_if_inconsistent(sql_base_types, files)
        if err_msg and "VARCHAR" not in sql_base_types:
            err_msgs.append(
                f"All sql_types for column '{col_name}' must be the same, "
                f"however --\n{err_msg}"
            )
            continue
        if err_msg:
            # If the sql_types are inconsistent and at least one of them
            # is a VARCHAR, then simply use VARCHAR
            # for this column across all files.
            sql_type = "VARCHAR"
            # The one single file that originally had sql_type as VARCHAR
            # has the detected "first row is headers" to be false
            # (i.e., the detection thought the first row was first row of data,
            # because it can't reliably tell when it comes to VARCHAR),
            # whereas for all the other files where the sql_type for this
            # column isn't VARCHAR, the detected "first row is headers" is
            # true instead. Due to this expected inconsistency,
            # we need to pass a flag out of this function to signal
            # that we can allow `headers` to be True.
            allow_inconsistent_headers = True
        else:
            sql_type = cols[0]["sql_type"]
        table_columns.append({"name": col_name, "sql_type": sql_type})

    if err_msgs:
        raise CivisImportError("\n".join(err_msgs))

    return table_columns, allow_inconsistent_headers


def _err_msg_if_inconsistent(items: list, files: list[_File]):
    if len(set(items)) <= 1:
        return
    values_to_indices = collections.defaultdict(list)
    for i, value in enumerate(items):
        values_to_indices[value].append(i)
    msg_for_each_value = [
        f"\t{v} from: {_format_files_for_err_msg(files, indices)}"
        for v, indices in values_to_indices.items()
    ]
    err_msg = "\n".join(msg_for_each_value)
    return err_msg


def _format_files_for_err_msg(files: list[_File], indices):
    files_in_str = [f"file {files[i].id} ({files[i].name})" for i in indices]
    return ", ".join(files_in_str)
